# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import numpy as np
from qiskit_metal import draw, Dict
from .base import QComponent
from numpy.linalg import norm
from typing import List, Tuple, Union, AnyStr
from collections.abc import Mapping
from qiskit_metal.toolbox_metal import math_and_overrides as mao
import math
import re


class QRoutePoint:
    """A convenience wrapper class to define an point with orientation, with a
    2D position and a 2D direction (XY plane).

    All values stored as np.ndarray of parsed floats.
    """

    def __init__(self, position: np.array, direction: np.array = None):
        """
        Args:
            position (np.ndarray of 2 points): Center point of the pin.
            direction (np.ndarray of 2 points): *Normal vector*
                This is the normal vector to the surface on which the pin mates.
                Defines which way it points outward. Has unit norm.  Defaults to None.
        """
        self.position = position
        if isinstance(position, list):
            if len(position[-1]) == 2:
                self.position = position[-1]
        self.direction = direction

    def __str__(self):
        return "position: " + str(self.position) + " : direction:" + str(
            self.direction) + "."


class QRoute(QComponent):
    """Super-class implementing routing methods that are valid irrespective of
    the number of pins (>=1). The route is stored in a n array of planar points
    (x,y coordinates) and one direction, which is that of the last point in the
    array. Values are stored as np.ndarray of parsed floats or np.array float
    pair.

    Inherits `QComponent` class

    Default Options:
        * pin_inputs: Dict
            * start_pin: Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to start from, which has a pin
                * pin: '' -- Name of pin used for pin_start
            * end_pin=Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to end on, which has a pin
                * pin: '' -- Name of pin used for pin_end
        * lead: Dict
            * start_straight: '0mm' -- Lead-in, defined as the straight segment extension from start_pin.  Defaults to 0.1um.
            * end_straight: '0mm' -- Lead-out, defined as the straight segment extension from end_pin.  Defaults to 0.1um.
            * start_jogged_extension: '' -- Lead-in, jogged extension of lead-in. Described as list of tuples
            * end_jogged_extension: '' -- Lead-out, jogged extension of lead-out. Described as list of tuples
        * fillet: '0'
        * total_length: '7mm'
        * trace_width: 'cpw_width' -- Defines the width of the line.  Defaults to 'cpw_width'.

    How to specify \*_jogged_extensions for the QRouteLeads:
        \*_jogged_extensions have to be specified in an OrderedDict with incremental keys.
        the value of each key specifies the direction of the jog and the extension past the jog.
        For example:

        .. code-block:: python
            :linenos:

            jogs = OrderedDict()
            jogs[0] = ["R", '200um']
            jogs[1] = ["R", '200um']
            jogs[2] = ["L", '200um']
            jogs[3] = ["L", '500um']
            jogs[4] = ["R", '200um']
            jogs_other = ....
            
            options = {'lead': {
                'start_straight': '0.3mm',
                'end_straight': '0.3mm',
                'start_jogged_extension': jogs,
                'end_jogged_extension': jogs_other
            }}

        The jog direction can be specified in several ways. Feel free to pick the one more
        convenient for your coding style:

        >> "L", "L#", "R", "R#", #, "#", "A,#", "left", "left#", "right", "right#"

        where # is any signed or unsigned integer or floating point value.
        For example the following will all lead to the same turn:

        >> "L", "L90", "R-90", 90, "90", "A,90", "left", "left90", "right-90"
    """

    component_metadata = Dict(short_name='route', _qgeometry_table_path='True')
    """Component metadata"""

    default_options = Dict(
        pin_inputs=Dict(
            start_pin=Dict(  # QRoute also supports single pin routes
                component='',  # Name of component to start from, which has a pin
                pin=''),  # Name of pin used for pin_start
            end_pin=Dict(
                component='',  # Name of component to end on, which has a pin
                pin='')  # Name of pin used for pin_end
        ),
        fillet='0',
        lead=Dict(start_straight='0mm',
                  end_straight='0mm',
                  start_jogged_extension='',
                  end_jogged_extension=''),
        total_length='7mm',
        trace_width='cpw_width')
    """Default options"""

    TOOLTIP = """QRoute"""

    def __init__(self,
                 design,
                 name: str = None,
                 options: Dict = None,
                 type: str = "CPW",
                 **kwargs):
        """Initializes all Routes.

        Calls the QComponent __init__() to create a new Metal component.
        Before that, it adds the variables that are needed to support routing.

        Args:
            design (QDesign): The parent design.
            name (str): Name of the component. Auto-named if possible.
            options (dict): User options that will override the defaults.  Defaults to None.
            type (string): Supports Route (single layer trace) and CPW (adds the gap around it).
                Defaults to "CPW".
        """
        # Class key Attributes:
        #     * head (QRouteLead()): Stores sequential points to start the route.
        #     * tail (QRouteLead()): (optional) Stores sequential points to terminate the route.
        #     * intermediate_pts: (list or numpy Nx2 or dict) Sequence of points between and other
        #         than head and tail.  Defaults to None. Type could be either list or numpy Nx2,
        #         or dict/OrderedDict nesting lists or numpy Nx2.
        #     * start_pin_name (string): Head pin name.  Defaults to "start".
        #     * end_pin_name (string): Tail pin name.  Defaults to "end".

        self.head = QRouteLead()
        self.tail = QRouteLead()

        # keep track of all points so far in the route from both ends
        self.intermediate_pts = np.empty((0, 2), float)  # will be numpy Nx2

        # supported pin names (constants)
        self.start_pin_name = "start"
        self.end_pin_name = "end"

        self.type = type.upper().strip()

        # # add default_options that are QRoute type specific:
        options = self._add_route_specific_options(options)

        # regular QComponent boot, including the run of make()
        super().__init__(design, name, options, **kwargs)

    def _add_route_specific_options(self, options):
        """Enriches the default_options to support different types of route
        styles.

        Args:
            options (dict): User options that will override the defaults

        Return:
            A modified options dictionary

        Raises:
            Exception: Unsupported route type
        """
        if self.type == "ROUTE":
            # all the defaults are fine as-is
            None
        elif self.type == "CPW":
            # add the variable to define the space between the route and the ground plane
            cpw_options = Dict(trace_gap='cpw_gap')
            if options:
                if "trace_gap" not in options:
                    # user did not pass the trace_gap, so add it
                    options.update(cpw_options)
            else:
                # user did not pass custom options, so create it to add trace_gap
                options["options"] = cpw_options
        else:
            raise Exception("Unsupported Route type: " + self.type +
                            " The only supported types are CPW and route")

        return options

    def _get_connected_pin(self, pin_data: Dict):
        """Recovers a pin from the dictionary.

        Args:
            pin_data: dict {component: string, pin: string}

        Return:
            The actual pin object.
        """
        return self.design.components[pin_data.component].pins[pin_data.pin]

    def set_pin(self, name: str) -> QRoutePoint:
        """Defines the CPW pins and returns the pin coordinates and normal
        direction vector.

        Args:
            name: String (supported pin names are: start, end)

        Return:
            QRoutePoint: Last point (for now the single point) in the QRouteLead

        Raises:
            Exception: Ping name is not supported
        """
        # First define which pin/lead you intend to initialize
        if name == self.start_pin_name:
            options_pin = self.options.pin_inputs.start_pin
            lead = self.head
        elif name == self.end_pin_name:
            options_pin = self.options.pin_inputs.end_pin
            lead = self.tail
        else:
            raise Exception("Pin name \"" + name +
                            "\" is not supported for this CPW." +
                            " The only supported pins are: start, end.")

        # grab the reference component pin
        reference_pin = self._get_connected_pin(options_pin)

        # create the cpw pin and document the connections to the reference_pin in the netlist
        self.add_pin(name, reference_pin.points[::-1], self.p.trace_width)
        self.design.connect_pins(
            self.design.components[options_pin.component].id, options_pin.pin,
            self.id, name)

        # anchor the correct lead to the pin and return its position and direction
        return lead.seed_from_pin(reference_pin)

    def set_lead(self, name: str) -> QRoutePoint:
        """Defines the lead_extension by adding a point to the self.head/tail.

        Args:
            name: String (supported pin names are: start, end)

        Return:
            QRoutePoint: Last point in the QRouteLead (self.head/tail)

        Raises:
            Exception: Ping name is not supported
        """
        p = self.parse_options()

        # First define which lead you intend to modify
        if name == self.start_pin_name:
            options_lead = p.lead.start_straight
            lead = self.head
            jogged_lead = self.p.lead.start_jogged_extension
        elif name == self.end_pin_name:
            options_lead = p.lead.end_straight
            lead = self.tail
            jogged_lead = self.p.lead.end_jogged_extension
        else:
            raise Exception("Pin name \"" + name +
                            "\" is not supported for this CPW." +
                            " The only supported pins are: start, end.")

        # then change the lead by adding a point in the same direction of the seed pin
        # minimum lead, to be able to jog correctly
        lead_length = max(options_lead, self.p.trace_width / 2.0)
        lead.go_straight(lead_length)
        # then add all the jogged lead information
        if jogged_lead:
            self.set_lead_extension(name)  # consider merging with set_lead

        # return the last QRoutePoint of the lead
        return lead.get_tip()

    def set_lead_extension(self, name: str) -> QRoutePoint:
        """Defines the jogged lead_extension by adding a series of turns to the
        self.head/tail.

        Args:
            name: String (supported pin names are: start, end)

        Return:
            QRoutePoint: Last point in the QRouteLead (self.head/tail)

        Raises:
            Exception: Ping name is not supported
            Exception: Dictionary error
        """
        p = self.parse_options()
        # First define which lead you intend to modify
        if name == self.start_pin_name:
            options_lead = p.lead.start_jogged_extension
            lead = self.head
        elif name == self.end_pin_name:
            options_lead = p.lead.end_jogged_extension
            lead = self.tail
        else:
            raise Exception("Pin name \"" + name +
                            "\" is not supported for this CPW." +
                            " The only supported pins are: start, end.")

        # then change the lead by adding points
        for turn, length in options_lead.values():
            if isinstance(turn, (float, int)):
                # turn is a number indicating the angle
                lead.go_angle(length, turn)
            elif re.search(r'^[-+]?(\d+\.\d+|\d+)$', turn):
                # turn is a string of a number indicating the angle
                lead.go_angle(length, float(turn))
            elif turn in ("left", "L"):
                # implicit turn -90 degrees
                lead.go_left(length)
            elif turn in ("right", "R"):
                # implicit turn 90 degrees
                lead.go_right(length)
            elif turn in ("straight", "D", "S"):
                # implicit 0 degrees movement
                lead.go_straight(length)
            elif re.search(r'^(left|L)[-+]?(\d+\.\d+|\d+)$', turn):
                # left turn by the specified int/float degrees. can be signed
                angle = re.sub(r'^(left|L)', "", turn)
                lead.go_angle(length, float(angle))
            elif re.search(r'^(right|R)[-+]?(\d+\.\d+|\d+)$', turn):
                # right turn by the specified int/float degrees. can be signed
                angle = re.sub(r'^(right|R)', "", turn)
                lead.go_angle(length, -1 * float(angle))
            elif ('A' or 'angle') in turn:
                # turn by the specified int/float degrees. Positive numbers turn left.
                turn, angle = turn.split(',')
                lead.go_angle(length, float(angle))
            else:
                raise Exception(
                    f"\nThe input string {turn} is not supported. Please specify the jog turn "
                    "using one of the supported formats:\n\"L\", \"L#\", \"R\", \"R#\", #, "
                    "\"#\", \"A,#\", \"left\", \"left#\", \"right\", \"right#\""
                    "\nwhere # is any signed or unsigned integer or floating point value.\n"
                    "For example the following will all lead to the same turn:\n"
                    "\"L\", \"L90\", \"R-90\", 90, "
                    "\"90\", \"A,90\", \"left\", \"left90\", \"right-90\"")

        # return the last QRoutePoint of the lead
        return lead.get_tip()

    def _get_lead2pts_array(self, arr) -> Tuple:
        """Return the last "diff pts" of the array. If the array is one
        dimensional or has only identical points, return -1 for tip_pt_minus_1.

        Return:
            Tuple: Of two np.ndarray. the arrays could be -1 instead, if point not found
        """
        pt = pt_minus_1 = None
        if len(arr) == 1:
            pt = arr[0]
        elif len(arr) > 1:
            if not isinstance(arr, np.ndarray) and len(arr) == 2 and len(
                    arr[0]) == 1:
                # array 2,1
                pt = arr
            else:
                # array N,2
                pt = arr[-1]
                prev_id = -2
                pt_minus_1 = arr[prev_id]
                while (pt_minus_1 == pt).all() and prev_id > -len(arr):
                    prev_id -= 1
                    pt_minus_1 = arr[prev_id]
                if (pt_minus_1 == pt).all():
                    pt_minus_1 = None
        return pt, pt_minus_1

    def get_tip(self) -> QRoutePoint:
        """Access the last element in the QRouteLead.

        Return:
            QRoutePoint: Last point in the QRouteLead
            The values are numpy arrays with two float points each.
        """
        if self.intermediate_pts is None:
            # no points in between, so just grab the last point from the lead-in
            return self.head.get_tip()
        tip_pt = tip_pt_minus_1 = None
        if isinstance(self.intermediate_pts, list) or isinstance(
                self.intermediate_pts, np.ndarray):
            tip_pt, tip_pt_minus_1 = self._get_lead2pts_array(
                self.intermediate_pts)
        elif isinstance(self.intermediate_pts, Mapping):
            # then it is either a dict or a OrderedDict
            # this method relies on the keys to be numerical integer. Will use the last points
            # assumes that the "value" associated with each key is some "not empty" list/array
            sorted_keys = sorted(self.intermediate_pts.keys(), reverse=True)
            for key in sorted_keys:
                pt0, pt_minus1 = self._get_lead2pts_array(
                    self.intermediate_pts[key])
                if pt0 is None:
                    continue
                if tip_pt_minus_1 is None:
                    tip_pt_minus_1 = pt0
                if tip_pt is None:
                    tip_pt, tip_pt_minus_1 = tip_pt_minus_1, tip_pt
                    tip_pt_minus_1 = pt_minus1
        else:
            print("unsupported type for self.intermediate_pts",
                  type(self.intermediate_pts))
            return
        if tip_pt is None:
            # no point in the intermediate array
            return self.head.get_tip()
        if tip_pt_minus_1 is None:
            # no "previous" point in the intermediate array
            tip_pt_minus_1 = self.head.get_tip().position

        return QRoutePoint(tip_pt, tip_pt - tip_pt_minus_1)

    def del_colinear_points(self, inarray):
        """Delete colinear points from the given array.

        Args:
            inarray (list): List of points

        Returns:
            list: List of points without colinear points
        """
        if len(inarray) <= 1:
            return
        else:
            outarray = list()  #outarray = np.empty(shape=[0, 2])
            pts = [None, None, inarray[0]]
            for idxnext in range(1, len(inarray)):
                pts = pts[1:] + [inarray[idxnext]]
                # delete identical points
                if np.allclose(*pts[1:]):
                    pts = [None] + pts[0:2]
                    continue
                # compare points once you have 3 unique points in pts
                if pts[0] is not None:
                    # if all(mao.round(i[1]) == mao.round(pts[0][1]) for i in pts) \
                    #         or all(mao.round(i[0]) == mao.round(pts[0][0]) for i in pts):
                    if mao.aligned_pts(pts):
                        pts = [None] + [pts[0]] + [pts[2]]
                # save a point once you successfully establish the three are not aligned,
                #  and before it gets dropped in the next loop cycle
                if pts[0] is not None:
                    outarray.append(pts[0])
            # save the remainder non-aligned points
            if pts[1] is not None:
                outarray.extend(pts[1:])
            else:
                outarray.append(pts[2])
            return np.array(outarray)

    def get_points(self) -> np.ndarray:
        """Assembles the list of points for the route by concatenating:
        head_pts + intermediate_pts, tail_pts.

        Returns:
            np.ndarray: ((H+N+T)x2) all points (x,y) of the CPW
        """
        # cover case where there is no intermediate points (straight connection between lead ends)
        if self.intermediate_pts is None:
            beginning = self.head.pts
        else:
            beginning = np.concatenate([self.head.pts, self.intermediate_pts],
                                       axis=0)

        # cover case where there is no tail defined (floating end)
        if self.tail is None:
            polished = beginning
        else:
            polished = np.concatenate([beginning, self.tail.pts[::-1]], axis=0)

        polished = self.del_colinear_points(polished)

        return polished

    def get_unit_vectors(self,
                         start: QRoutePoint,
                         end: QRoutePoint,
                         snap: bool = False) -> Tuple:
        """Return the unit and target vector in which the CPW should process as
        its coordinate sys.

        Args:
            start (QRoutePoint): Reference start point (direction from here)
            end (QRoutePoint): Reference end point (direction to here)
            snap (bool): True to snap to grid.  Defaults to False.

        Returns:
            array: straight and 90 deg CCW rotated vecs 2D
            (array([1., 0.]), array([0., 1.]))
        """
        # handle chase when start and end are same?
        v = end.position - start.position
        direction = v / norm(v)
        if snap:
            direction = draw.Vector.snap_unit_vector(direction, flip=False)
        normal = draw.Vector.rotate(direction, np.pi / 2)
        return direction, normal

    @property
    def length(self) -> float:
        """Sum of all segments length, including the head.

        Return:
            length (float): Full point_array length
        """
        # get the final points (get_point also eliminate co-linear and short edges)
        points = self.get_points()

        # get the length without the corner rounding radius adjustment
        length_estimate = sum(
            norm(points[i + 1] - points[i]) for i in range(len(points) - 1))
        # compensate for corner rounding
        length_estimate -= self.length_excess_corner_rounding(points)

        return length_estimate

    def length_excess_corner_rounding(self, points) -> float:
        """Computes how much length to deduce for compensating the fillet
        settings.

        Args:
            points (list or array): List of vertices that will be receiving the corner rounding radius

        Return:
            length_excess (float): Corner rounding radius excess multiplied by the number of points
        """
        # deduct the corner rounding (WARNING: assumes fixed fillet for all corners)
        length_arch = 0.5 * self.p.fillet * math.pi
        length_corner = 2 * self.p.fillet
        length_excess = length_corner - length_arch
        # the start and and point are the pins, so no corner rounding
        return (len(points) - 2) * length_excess

    def assign_direction_to_anchor(self, ref_pt: QRoutePoint,
                                   anchor_pt: QRoutePoint):
        """Method to assign a direction to a point. Currently assigned as the
        max(x,y projection) of the direct path between the reference point and
        the anchor. Method directly modifies the anchor_pt.direction, thus
        there is no return value.

        Args:
            ref_pt (QRoutePoint): Reference point
            anchor_pt (QRoutePoint): Anchor point. if it already has a direction, the method will not overwrite it
        """
        if anchor_pt.direction is not None:
            # anchor_pt already has a direction (not an anchor?), so do nothing
            return
        # Current rule: stop_direction aligned with longer edge of the rectangle connecting ref_pt and anchor_pt
        ref = ref_pt.position
        anchor = anchor_pt.position
        # Absolute value of displacement between ref and anchor in x direction
        offsetx = abs(anchor[0] - ref[0])
        # Absolute value of displacement between ref and anchor in y direction
        offsety = abs(anchor[1] - ref[1])
        if offsetx >= offsety:  # "Wide" rectangle -> anchor_arrow points along x
            assigned_direction = np.array([ref[0] - anchor[0], 0])
        else:  # "Tall" rectangle -> anchor_arrow points along y
            assigned_direction = np.array([0, ref[1] - anchor[1]])
        anchor_pt.direction = assigned_direction / norm(assigned_direction)

    def make_elements(self, pts: np.ndarray):
        """Turns the CPW points into design elements, and add them to the
        design object.

        Args:
            pts (np.ndarray): Array of points
        """

        # prepare the routing track
        line = draw.LineString(pts)

        # compute actual final length
        p = self.p
        self.options._actual_length = str(
            line.length - self.length_excess_corner_rounding(line.coords)
        ) + ' ' + self.design.get_units()

        # expand the routing track to form the substrate core of the cpw
        self.add_qgeometry('path', {'trace': line},
                           width=p.trace_width,
                           fillet=p.fillet,
                           layer=p.layer)
        if self.type == "CPW":
            # expand the routing track to form the two gaps in the substrate
            # final gap will be form by this minus the trace above
            self.add_qgeometry('path', {'cut': line},
                               width=p.trace_width + 2 * p.trace_gap,
                               fillet=p.fillet,
                               layer=p.layer,
                               subtract=True)


class QRouteLead:
    """A simple class to define a an array of points with some properties,
    defines 2D positions and some of the 2D directions (XY plane).

    All values stored as np.ndarray of parsed floats.
    """

    def __init__(self, *args, **kwargs):
        """QRouteLead is a simple sequence of points.

        Used to accurately control one of the QRoute termination points
        Before that, it adds the variables that are needed to support routing.

        Attributes:
            pts (numpy Nx2): Sequence of points.  Defaults to None.
            direction (numpy 2x1): Normal from the last point of the array.  Defaults to None.
        """
        # keep track of all points so far in the route from both ends
        self.pts = None  # will be numpy Nx2
        # keep track of the direction of the tip of the lead (last point)
        self.direction = None  # will be numpy 2x1

    def seed_from_pin(self, pin: Dict) -> QRoutePoint:
        """Initialize the QRouteLead by giving it a starting point and a
        direction.

        Args:
            pin: object describing the "reference_pin" (not cpw_pin) this is attached to.
                this is currently (8/4/2020) a dictionary

        Return:
            QRoutePoint: Last point (for now the single point) in the QRouteLead
            The values are numpy arrays with two float points each.
        """
        position = pin['middle']
        direction = pin['normal']

        self.direction = direction
        self.pts = np.array([position])

        return QRoutePoint(position, direction)

    def go_straight(self, length: float):
        """Add a point ot 'length' distance in the same direction.

        Args:
            length (float) : How much to move by
        """
        self.pts = np.append(self.pts, [self.pts[-1] + self.direction * length],
                             axis=0)

    def go_left(self, length: float):
        """Straight line 90deg counter-clock-wise direction w.r.t. lead tip
        direction.

        Args:
            length (float): How much to move by
        """
        self.direction = draw.Vector.rotate(self.direction, np.pi / 2)
        self.pts = np.append(self.pts, [self.pts[-1] + self.direction * length],
                             axis=0)

    def go_right(self, length: float):
        """Straight line 90deg clock-wise direction w.r.t. lead tip direction.

        Args:
            length (float): How much to move by
        """
        self.direction = draw.Vector.rotate(self.direction, -1 * np.pi / 2)
        self.pts = np.append(self.pts, [self.pts[-1] + self.direction * length],
                             axis=0)

    def go_right45(self, length: float):
        """Straight line at 45 angle clockwise w.r.t lead tip direction.

        Args:
            length(float): How much to move by
        """
        self.direction = draw.Vector.rotate(self.direction, -1 * np.pi / 4)
        self.pts = np.append(self.pts, [self.pts[-1] + self.direction * length],
                             axis=0)

    def go_left45(self, length: float):
        """Straight line at 45 angle counter-clockwise w.r.t lead tip direction.

        Args:
            length(float): How much to move by
        """
        self.direction = draw.Vector.rotate(self.direction, np.pi / 4)
        self.pts = np.append(self.pts, [self.pts[-1] + self.direction * length],
                             axis=0)

    def go_angle(self, length: float, angle: float):
        """ Straight line at any angle w.r.t lead tip direction.

        Args:
            length(float): How much to move by
            angle(float): rotation angle w.r.t lead tip direction
        """
        self.direction = draw.Vector.rotate(self.direction, np.pi / 180 * angle)
        self.pts = np.append(self.pts, [self.pts[-1] + self.direction * length],
                             axis=0)

    @property
    def length(self):
        """Sum of all segments length, including the head.

        Return:
            length (float): Full point_array length
        """
        return sum(
            norm(self.pts[i + 1] - self.pts[i])
            for i in range(len(self.pts) - 1))

    def get_tip(self) -> QRoutePoint:
        """Access the last element in the QRouteLead.

        Return:
            QRoutePoint: Last point in the QRouteLead
            The values are numpy arrays with two float points each.
        """
        if self.pts.ndim == 1:
            return QRoutePoint(self.pts, self.direction)
        return QRoutePoint(self.pts[-1], self.direction)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This is the main module that defines what an element is in Qiskit Metal.

See the docstring of `QGeometryTables`
"""

import inspect
import logging
import pandas as pd
import shapely

from typing import TYPE_CHECKING
from typing import Dict as Dict_
from typing import List, Tuple, Union, Any, Iterable
from geopandas import GeoDataFrame, GeoSeries

from .. import Dict
from ..draw import BaseGeometry
from qiskit_metal.draw.utility import round_coordinate_sequence

from shapely.geometry.multipolygon import MultiPolygon  #to avoid MultiPolygons
from .. import config
if not config.is_building_docs():
    from qiskit_metal.toolbox_python.utility_functions import get_range_of_vertex_to_not_fillet, data_frame_empty_typed

if TYPE_CHECKING:
    from ..qlibrary.core import QComponent
    from ..designs import QDesign

__all__ = ['is_qgeometry_table', 'QGeometryTables']  # , 'ElementTypes']

# from collections import OrderedDict
# dict are ordered in Python 3.6+ by default, this is for backward compatibility

# class ElementTypes:
#     """
#     Types of qgeometry
#         positive : qgeometry that are positive mask
#         negative : qgeometry that will be subtracted from the chip ground plane
#         helper   : qgeometry that are only used in a helper capacity,
#                    such as labels or mesh rectangles
#     """
#     positive = 0
#     negative = 1
#     helper   = 2


def is_qgeometry_table(obj):
    """Check if an object is a Metal BaseElementTable, i.e., an instance of
    `QGeometryTables`.

    The problem is that the `isinstance` built-in method fails
    when this module is reloaded.

    Args:
        obj (object): Test this object

    Returns:
        bool: True if is a Metal element
    """
    if isinstance(obj, Dict):
        return False

    return hasattr(obj, '__i_am_qgeometry_table__')


#############################################################################
#
# Dictionary that specifies the column names of various element tables.
#

# TODO: implement data types in construction of tables?
# TODO: when a copy over the table and manipulate ehtmee
# i seem to loose the assignments to bool etc.

ELEMENT_COLUMNS = dict(

    ################################################
    # DO NOT MODIFY THE base DICTIONARY.
    # This is for Metal API use only.
    # To add a new element type, add a new key below.
    base=dict(
        component=str,  # Unique ID of the component to which the element belongs
        name=str,  # name of the element
        geometry=object,  # shapely object
        layer=int,  # gds type of layer
        subtract=bool,  # do we subtract from the ground place of the chip
        helper=bool,  # helper or not
        chip=str,  # chip name
        # type=str,  # metal, helper.   poly=10 or path=11
        __renderers__=dict(
            # ADD specific renderers here, all renderes must register here.
            # hfss = dict( ... ) # pick names as hfss_name
            # hfss=dict(
            #     boundary_name=str,
            #     material=str,
            #     perfectE=bool
            # ),
            # gds=dict(
            #     type=str,
            #     color=str,
            # )
        )),

    ################################################
    # Specifies a path, such as a CPW.
    # Ideas: chamfer
    path=dict(
        width=float,
        fillet=object,  # TODO: not decided yet how to represent this
        __renderers__=dict()),

    ################################################
    # Specifies a polygon
    # Ideas: chamfer
    poly=dict(
        fillet=object,  # TODO: not decided yet how to represent this
        __renderers__=dict()),
    ################################################
    # Specifies a junction as a 2 point line and width
    # This should provide enough information so as to
    # - render a sheet with inductance (from renderer options) + a vector for EPR
    # - generate ports (edge ports?) for Z analysis
    # - provice bounding box dimensions for the p-cell of ebeam junction layout
    #       for GDS renderer
    junction=dict(width=float, __renderers__=dict()),
    ################################################
    # Specifies a curved object, such as a circle. Perhaps as a buffered point
    # Not yet implemented
    # curved = dict(
    # __renderers__= dict(
    # )
    # )
)
"""Dictionary that specifies the column names of various element tables."""

#############################################################################
#
# Class to create, store, and handle element tables.
#
TRUE_BOOLS = [True, 'True', 'true', 'Yes', 'yes', '1', 1]


class QGeometryTables(object):
    """Class to create, store, and handle element tables.

    A regular user would not need to create tables themselves.
    This is handled automatically by the design creation and plugins.

    Structure:
        A component, such as a qubit, is a collection of qgeometry.
        For example, an element includes a rectangle, a cpw path, or a more general polygon.

        An element is a row in a table.

        All qgeometry of a type (Path or Polygon, or otherwise) are stored in a
        single table of their element type.

        All qgeometry of the same kind are stored in a table.
        A renderer has to know how to handle all types of qgeometry in order to render them.

    For plugin developers:
        In the following, we provide an example that illustrates for plugin developers how
        to add custom qgeometry and custom element properties. For example, we will add, for a renderer
        called hfss, a string property called 'boundary', a bool property called 'perfectE', and a property called 'material'.

    For plugin developers, example use:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 4,6

            import qiskit_metal as metal

            design = metal.designs.DesignPlanar()
            design.qgeometry = metal.QGeometryTables(design)

            # return the path table - give access to ..
            design.qgeometry['path']
            design.qgeometry.table['path']

            # Define interfaces
            design.qgeometry.get_component(
                    component_name,
                    element_name,
                    columns=all or geom or list) # get all elements for components

            >>> component	name	geometry	layer	type	chip	subtract	fillet	color	width


    Now, if we want to add custom qgeometry through two fake renderers called hfss and gds:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 1-15

            metal.QGeometryTables.add_renderer_extension('hfss', dict(
                base=dict(
                    boundary=str,
                    perfectE=bool,
                    material=str,
                    )
                ))

            metal.QGeometryTables.add_renderer_extension('gds', dict(
                path=dict(
                    color=str,
                    pcell=bool,
                    )
                ))

            design = metal.designs.DesignPlanar()
            qgeometry = metal.QGeometryTables(design)

            qgeometry.tables['path']
            >>> component	name	geometry	layer	type	chip	subtract	fillet	color	width	hfss_boundary	hfss_perfectE	hfss_material	gds_color	gds_pcell
    """

    # Dummy private attribute used to check if an instanciated object is
    # indeed a element table class. The problem is that the `isinstance`
    # built-in method fails when this module is reloaded.
    # Used by `is_element` to check.
    __i_am_qgeometry_table__ = True

    ELEMENT_COLUMNS = ELEMENT_COLUMNS
    """
    Table column names to use to create.
    This dict should be updated by renderers.
    """

    # For creating names of columns of renderer properties
    name_delimiter = '_'
    """ Delimiter to use when creating names of columns of renderer properties. """

    def __init__(self, design: 'QDesign'):
        """The constructor for the `QGeometryTables` class.

        Args:
            design: Design in use
        """
        self._design = design

        self._tables = Dict()

        # Need to call after columns are added by add_renderer_extension is run by all the renderers.
        # self.create_tables()

    @property
    def design(self) -> 'QDesign':
        """Return a reference to the parent design object."""
        return self._design

    @property
    def logger(self) -> logging.Logger:
        """Return the logger."""
        return self._design.logger

    @property
    def tables(self) -> Dict_[str, GeoDataFrame]:
        """The dictionary of tables containing qgeometry.

        Returns:
            Dict_[str, GeoDataFrame]: The keys of this dictionary are
            also obtained from `self.get_element_types()`
        """
        return self._tables

    @classmethod
    def add_renderer_extension(cls, renderer_name: str, qgeometry: dict):
        """Add renderer element extension to ELEMENT_COLUMNS. Called when the
        load function of a renderer is called.

        Args:
            renderer_name (str): Name of renderer
            qgeometry (dict):  dict of dict.  Keys give element type names,
                              such as base, poly, path, etc.
        """
        # Make sure that the base and all other element kinds have this renderer registered
        for element_key in cls.ELEMENT_COLUMNS:
            if not renderer_name in cls.ELEMENT_COLUMNS[element_key][
                    '__renderers__']:
                cls.ELEMENT_COLUMNS[element_key]['__renderers__'][
                    renderer_name] = dict()

        # Now update the dictionaries with all qgeometry that the renderer may have
        for element_key, element_column_ext_dict in qgeometry.items():

            # The element the render is specifying is not in the specified qgeometry;
            # then add it. This shouldn't really happen.
            # The rest of the renderer dict keys in __renderers__  are missing for
            # the created type. Avoid doing, else hope it works.
            if not element_key in cls.ELEMENT_COLUMNS:
                cls.ELEMENT_COLUMNS[element_key] = dict(__renderers__=dict())

            # Now add qgeometry
            cls.ELEMENT_COLUMNS[element_key]['__renderers__'][
                renderer_name].update(element_column_ext_dict)

    # could use weakref memorization
    # https://stackoverflow.com/questions/33672412/python-functools-lru-cache-with-class-methods-release-object
    @classmethod
    def get_element_types(cls) -> List[str]:
        """Return the names of the available qgeometry to create. This does not
        include 'base', but is rather such as poly and path.

        Returns:
            list(str) : A list of name in self.ELEMENT_COLUMNS
        """
        # TODO: I should probably make this a variable and memorize, only change when qgeometry are added and removed
        # can be slow for performance to look up each time and recalculate, since may call this often
        names = list(cls.ELEMENT_COLUMNS.keys())
        names.remove('base')
        return names

    def create_tables(self):
        """Creates the default tables once. Populates the dict 'tables' of
        GeoDataFrame, each with columns corresponding to the types of qgeometry
        defined in ELEMENT_COLUMNS.

        Should only be done once when a new design is created.
        """
        self.logger.debug('Creating Element Tables.')

        for table_name in self.get_element_types():
            # Create GeoDataFrame with correct columns and d types
            assert isinstance(table_name, str)
            assert table_name.isidentifier()

            # Get column names
            # Base names, add concrete names, then add renderer names

            # Base names
            columns_base = self.ELEMENT_COLUMNS['base'].copy()
            columns_base_renderers = columns_base.pop('__renderers__')

            # Concrete names
            columns_concrete = self.ELEMENT_COLUMNS[table_name].copy()
            columns_concrete_renderer = columns_concrete.pop('__renderers__')

            assert isinstance(columns_base_renderers, dict) and\
                isinstance(columns_concrete_renderer, dict),\
                "Please make sure that all qgeometry types have __renderers__\
                     which is a dictionary."

            # Combine all base names and renderer names
            columns = columns_base
            columns.update(columns_concrete)
            # add renderer columns: base and then concrete
            for renderer_key in columns_base_renderers:
                columns.update(
                    self._prepend_renderer_names(table_name, renderer_key,
                                                 columns_base_renderers))
                columns.update(
                    self._prepend_renderer_names(table_name, renderer_key,
                                                 columns_concrete_renderer))

            # Validate -- Throws an error if not valid
            self._validate_column_dictionary(table_name, columns)

            # Create df with correct column names
            table = GeoDataFrame(data_frame_empty_typed(columns))
            # not used elsewhere, also the name becomes "name" for some reason
            table.name = table_name

            # Assign
            self.tables[table_name] = table

    def _validate_column_dictionary(self, table_name: str, column_dict: dict):
        """Validate A possible error here is if the user did not pass a valid
        data type.

        Throws an error if not valid.

        Args:
            table_name (str): Name of element table (e.g., 'poly')
            column_dict (dict): Dictionary to check

        Raises:
            TypeError: Data type '' not understood
        """
        __pre = 'ERROR CREATING ELEMENT TABLE FOR DESIGN: \
            \n  ELEMENT_TABLE_NAME = {table_name}\
            \n  KEY                = {k} \
            \n  VALUE              = {v}\n '

        # Are these assertions still holding true?
        for k, v in column_dict.items():
            assert isinstance(k, str), __pre.format(**locals()) +\
                ' Key needs to be a string!'
            assert k.isidentifier(), __pre.format(**locals()) +\
                ' Key needs to be a valid string identifier!'
            assert inspect.isclass(v), __pre.format(**locals()) +\
                ' Value needs to be a class!'

    def get_rname(self, renderer_name: str, key: str) -> str:
        """Get name for renderer property.

        Args:
            renderer_name (str): Name of the renderer
            key (str): Key to get the name for

        Returns:
            str: The unique named used as a column in the table
        """
        return renderer_name + self.name_delimiter + key

    def _prepend_renderer_names(self, table_name: str, renderer_key: str,
                                rdict: dict):
        """Prepare all the renderer names.

        Args:
            table_name (str): Unused
            renderer_key (str): Key to check for
            rdict (dict): Renderer dictionary

        Returns:
            dict: Prepared dictionary

        TODO:
            This function has arguments that are unused, fix the function or ditch the unused args
        """
        return {
            self.get_rname(renderer_key, k): v
            for k, v in rdict.get(renderer_key, {}).items()
        }

    def add_qgeometry(
            self,
            kind: str,
            component_name: str,
            geometry: dict,
            subtract: bool = False,
            helper: bool = False,
            layer: Union[int, str] = 1,  # chip will be here
            chip: str = 'main',
            **other_options):
        """Main interface to add qgeometries.

        Args:
            kind (str): Must be in get_element_types ('path', 'poly', etc.).
            component_name (str): Component name.
            geometry (dict): Dict of shapely geometry.
            subtract (bool): Subtract - passed through.  Defaults to False.
            helper (bool): Helper - passed through.  Defaults to False.
            layer (Union[int, str]): Layer - passed through.  Defaults to 1.
            chip (str): Chip name - passed through.  Defaults to 'main'.
            **other_options (object): Other_options - passed through.
        """
        # TODO: Add unit test

        # ensure correct types
        if not isinstance(subtract, bool):
            subtract = subtract in TRUE_BOOLS
        if not isinstance(helper, bool):
            helper = helper in TRUE_BOOLS

        if not (kind in self.get_element_types()):
            self.logger.error(
                f'Creator user error: Unknown element kind=`{kind}`'
                f'Kind must be in {self.get_element_types()}. This failed for component'
                f'name = `{component_name}`.\n'
                f' The call was with subtract={subtract} and helper={helper}'
                f' and layer={layer}, and options={other_options}')

        #Checks if (any) of the geometry are MultiPolygons, and breaks them up into
        #individual polygons. Rounds the coordinate sequences of those values to avoid
        #numerical errors.
        rounding_val = self.design.template_options['PRECISION']
        new_dict = Dict()
        for key, item in geometry.items():
            if isinstance(geometry[key], MultiPolygon):
                temp_multi = geometry[key]
                shape_count = 0
                for shape_temp in temp_multi.geoms:
                    new_dict[key + '_' +
                             str(shape_count)] = round_coordinate_sequence(
                                 shape_temp, rounding_val)
                    shape_count += 1
            else:
                new_dict[key] = round_coordinate_sequence(item, rounding_val)

        geometry = new_dict

        # Create options TODO: Might want to modify this (component_name -> component_id)
        # Give warning if length is to be fillet's and not long enough.
        self.check_lengths(geometry, kind, component_name, **other_options)

        # Create options
        options = dict(component=component_name,
                       subtract=subtract,
                       helper=helper,
                       layer=int(layer),
                       chip=chip,
                       **other_options)

        #replaces line above to generate the options.
        #for keyC in design.qgeometry.tables[kind].columns:
        #    if keyC != 'geometry':
        #        options[keyC] = ???[keyC] -> alternative manner to pass options to the add_qgeometry function?
        #                                       instead have the add_qeometry in baseComponent generate the dict?

        #Could we just append rather than make a new table each time? This seems slow
        table = self.tables[kind]

        # assert that all names in options are in table columns! TODO: New approach will not be wanting
        #to do this (maybe check that all columns are in options?)
        df = GeoDataFrame.from_dict(geometry,
                                    orient='index',
                                    columns=['geometry'])
        df.index.name = 'name'
        df = df.reset_index()

        df = df.assign(**options)

        # Set new table. Unfortunately, this creates a new instance. Can just direct append
        self.tables[kind] = pd.concat([table, df],
                                      axis=0,
                                      join='outer',
                                      ignore_index=True,
                                      sort=False,
                                      verify_integrity=False,
                                      copy=False)

    def check_lengths(self, geometry: shapely.geometry.base.BaseGeometry,
                      kind: str, component_name: str, **other_options):
        """If user wants to fillet, check the line-segments to see if it is too
        short for fillet.

        Args:
            geometry (shapely.geometry.base.BaseGeometry): The LineString to investigate.
            kind (str): Name of table, i.e. 'path', 'poly', 'junction, etc
            component_name (str): Is an integer id.
        """

        if 'fillet' in other_options.keys():

            fillet = other_options['fillet']

            for key, geom in geometry.items():
                if isinstance(geom, shapely.geometry.LineString):
                    coords = list(geom.coords)
                    qdesign_precision = self.design.template_options.PRECISION
                    range_vertex_of_short_segments = get_range_of_vertex_to_not_fillet(
                        coords, fillet, qdesign_precision, add_endpoints=False)

                    if len(range_vertex_of_short_segments) > 0:
                        range_string = ""
                        for item in range_vertex_of_short_segments:

                            range_string += f'({ item[0]}-{item[1]}) '
                        text_id = self.design._components[component_name]._name
                        self.logger.warning(
                            f'For {kind} table, component={text_id}, key={key}'
                            f' has short segments that could cause issues with fillet. Values in {range_string} '
                            f'are index(es) in shapely geometry.')

    def parse_value(self, value: Union[Any, List, Dict, Iterable]) -> Any:
        """Same as design.parse_value. See design for help.

        Returns:
            Parsed value of input.
        """
        return self.design.parse_value(value)

    def clear_all_tables(self):
        """Clear all the internal tables and all else.

        Use when clearing a design and starting from scratch.
        """
        self.tables.clear()
        self.create_tables()  # remake all tables

    def delete_component(self, name: str):
        """Delete component by name.

        Args:
            name (str): Name of component (case sensitive)
        """
        # TODO: Add unit test
        # TODO: is this the best way to do this, or is there a faster way?
        a_comp = self.design.components[name]
        if a_comp is not None:
            for table_name in self.tables:
                df = self.tables[table_name]
                self.tables[table_name] = df[df['component'] != a_comp.id]

    def delete_component_id(self, component_id: int):
        """Drop the components within the qgeometry.tables.

        Args:
            component_id (int): Unique number to describe the component.
        """
        for table_name in self.tables:
            df_table_name = self.tables[table_name]
            # self.tables[table_name] = df_table_name.drop(df_table_name[df_table_name['component'] == component_id].index)
            self.tables[table_name] = df_table_name[df_table_name['component']
                                                    != component_id]

    def get_component(
        self,
        name: str,
        table_name: str = 'all'
    ) -> Union[GeoDataFrame, Dict_[str, GeoDataFrame]]:
        """Return the table for just a given component. If all, returns a
        dictionary with keys as table names and tables of components as values.

        Args:
            name (str): Name of component (case sensitive).  Defaults to 'all'.
            table_name (str): Element table name ('poly', 'path', etc.).  Defaults to {'all'}.

        Returns:
            Union[GeoDataFrame, Dict_[str, GeoDataFrame]]: Either a GeoDataFrame or a dict or GeoDataFrame.

        Example usage:
            ```table = pd.concat(qgeometry.get_component('Q1')) # , axis=0```
        """

        if table_name == 'all':
            tables = {}
            for table_name in self.get_element_types():
                tables[table_name] = self.get_component(name, table_name)
            return tables
        else:
            df = self.tables[table_name]
            a_comp = self.design.components[name]
            if a_comp is None:
                # Component not found.
                return None
            else:
                return df[df.component == a_comp.id]

            # comp_id = self.design.components[name].id
            # return df[df.component == comp_id]

    def get_component_bounds(self,
                             name: str) -> Tuple[float, float, float, float]:
        """Returns a tuple containing minx, miny, maxx, maxy values for the
        bounds of the component as a whole.

        Args:
            name (str): Component name

        Returns:
            Geometry: Bare element geometry
        """
        gs = self.get_component_geometry(name)  # Pandas GeoSeries
        if len(gs) == 0:
            return (0, 0, 0, 0)
        else:
            return gs.total_bounds

    def rename_component(self, component_id: int, new_name: str):
        """Rename component by ID (integer) cast to string format.

        Args:
            component_id (int) : ID of component (case sensitive)
            new_name (str) : The new name of the component (case sensitive)
        """

        # comp_id = self.design.components[name].id
        component_int_id = int(component_id)
        a_comp = self.design._components[component_int_id]
        if a_comp is None:
            return None
        else:
            # TODO: is this the best way to do this, or is there a faster way?
            for table_name in self.tables:
                table = self.tables[table_name]
                table.component[table.component == a_comp.id] = new_name

    def get_component_geometry_list(self,
                                    name: str,
                                    table_name: str = 'all'
                                   ) -> List[BaseGeometry]:
        """Return just the bare element geometry (shapely geometry objects) as
        a list, for the selected component.

        Args:
            name (str) : Name of component (case sensitive)
            table_name (str) : Element type ('poly', 'path', etc.).
                               Can also be 'all' to return all. This is the default.

        Returns:
            list: List of shapley.geometry objects
        """
        if table_name == 'all':
            qgeometry = []
            for table in self.get_element_types():
                qgeometry += self.get_component_geometry_list(name, table)

        else:
            table = self.tables[table_name]
            comp_id = self.design.components[name].id
            qgeometry = table.geometry[table.component == comp_id].to_list()

        return qgeometry

    def get_component_geometry(self, name: str) -> GeoSeries:
        """Returns geometry of a given component.

        Args:
            name (str) : Name of component (case sensitive)

        Returns:
            GeoSeries : Geometry of the component
        """
        comp_id = self.design.components[name].id
        qgeometry = {}
        for table_name in self.get_element_types():
            table = self.tables[table_name]
            qgeometry[table_name] = table.geometry[table.component == comp_id]
        qgeometry = pd.concat(qgeometry)

        # when concatenating empty GeoSeries, returns Series (ugly fix)
        if not isinstance(qgeometry, GeoSeries):
            qgeometry = GeoSeries(qgeometry)

        return qgeometry

    def get_component_geometry_dict(self,
                                    name: str,
                                    table_name: str = 'all'
                                   ) -> List[BaseGeometry]:
        """Return just the bare element geometry (shapely geometry objects) as
        a dict, with key being the names of the qgeometry and the values as the
        shapely geometry, for the selected component.

        Args:
            name (str) : Name of component (case sensitive)
            table_name (str) : Element type ('poly', 'path', etc.). Defaults to 'all'.

        Returns:
            dict: Bare element geometry
        """
        if table_name == 'all':
            qgeometry = Dict()
            for table in self.get_element_types():
                qgeometry[table] = self.get_component_geometry_list(name, table)
            return qgeometry  # return pd.concat(qgeometry, axis=0)

        else:
            table = self.tables[table_name]

            # mask the rows nad get only 2 columns
            comp_id = self.design.components[name].id
            df_comp_id = table.loc[table.component == comp_id,
                                   ['name', 'geometry']]
            df_geometry = df_comp_id.geometry
            df_geometry.index = df_comp_id.name
            return df_geometry.to_dict()

    def check_element_type(self,
                           table_name: str,
                           log_issue: bool = True) -> bool:
        """Check if the name `table_name` is in the element tables.

        Args:
            table_name (str): Element type ('poly', 'path', etc.) or 'all'
            log_issue (bool): Throw an error in the log if name missing.  Defaults to True.

        Returns:
            bool: True if the name is valid, else False
        """
        if not table_name in self.get_element_types() or table_name in 'all':
            if log_issue:
                self.logger.error(
                    f'Element Tables: Tried to access non-existing element table: `{table_name}`'
                )
            return False
        else:
            return True

    def get_all_unique_layers_for_all_tables(self,
                                             qcomp_ids: Union[list, None] = None
                                            ) -> list:
        """Get a list of all unique layer number used in all of the geometry tables.
        User can get for all components or a subset.

        Args:
            qcomp_ids (Union[list, None], optional): The list has integers
                                        which denote component_id. Defaults to None.

        Returns:
            list: The unique layer numbers for the list of component ids passed in argument.
        """
        if qcomp_ids is None:
            qcomp_ids = []
        frames = list()
        unique_layers = None

        if len(qcomp_ids) == 0:
            # use all components
            frames = [self.tables[a_df] for a_df in self.tables]
        else:
            # Use just the component ID's in qcomp_ids.
            for table_key in self.tables:
                temp_df = self.tables[table_key]
                mask_temp = temp_df['component'].isin(qcomp_ids)
                subset_temp_df = temp_df[mask_temp]
                frames.append(subset_temp_df)

        #Concat the frames and then determine the unique layer numbers.
        unique_layers = list(
            pd.concat(frames, ignore_index=True)['layer'].unique())

        return unique_layers

    def get_all_unique_layers(self, chip_name: str) -> list:
        """Returns a lit of unique layers for the given chip names.

        Args:
            chip_name (str): Name of the chip

        Returns:
            list: List of unique layers
        """
        unique_layers = list()
        for table_name in self.design.qgeometry.get_element_types():
            table = self.design.qgeometry.tables[table_name]
            temp = table[table['chip'] == chip_name]
            layers = temp['layer'].unique().tolist()
            unique_layers += layers
        unique_layers = list(set(unique_layers))

        return unique_layers



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Plotting functions for shapely components using mpl."""

import matplotlib as mpl
import matplotlib.cbook as cbook
import matplotlib.image as image
import matplotlib.pyplot as plt
import numpy as np
import shapely
from cycler import cycler
from matplotlib.cbook import _OrderedSet
from shapely.geometry import LinearRing, Polygon  # Point, LineString,

from ... import Dict
from ...draw import BaseGeometry
from .mpl_interaction import figure_pz
from .patch import PolygonPatch

__all__ = [
    '_render_poly_zkm', 'render_poly', 'render', 'style_axis_simple',
    'get_prop_cycle', 'style_axis_standard', 'figure_spawn',
    '_axis_set_watermark_img', 'clear_axis'
]

##########################################################################################
# Plotting subroutines

# Todo Move - default config
style_config = Dict(
    poly=dict(lw=1, ec='k', alpha=0.5)  # fc='#6699cc',
    # Dict(
    #exterior = dict(lw=1, edgecolors='k', alpha=0.5),
    # interior = dict(facecolors='w', lw=1, edgecolors='grey'))
)
"""Style configuration"""


def _render_poly_zkm(poly: Polygon, ax, kw=None, kw_hole=None):
    """Style and draw a shapely polygon using MPL.

    Args:
        poly (Polygon): The polygon.
        ax (plt.Axes): Matplotlib axis to render to.
        kw (dict): Dictionary of kwargs for the plotting.  Defaults to None.
        kw_hold (kw_hole): kw_hole.  Defaults to None.
    """
    if kw_hole is None:
        kw_hole = {}
    if kw is None:
        kw = {}

    # should probably  only create these once and pass these in here
    kw = {**style_config.poly.exterior, **kw}
    kw_hole = {**style_config.poly.interior, **kw}

    if not poly.exterior == LinearRing():  # not empty - better check?
        # pylint: disable=protected-access
        # Exterior
        coords = np.array(poly.exterior.coords)
        mpl_poly = mpl.patches.Polygon(coords)
        color = ax._get_lines.get_next_color()
        ax.add_collection(
            mpl.collections.PatchCollection([mpl_poly], **{
                **{
                    'facecolor': color
                },
                **kw
            }))

        # Interior (i.e., holes)
        polys = []
        for hole in poly.interiors:
            coords = tuple(hole.coords)
            poly = mpl.patches.Polygon(coords)
            polys.append(poly)
        ax.add_collection(mpl.collections.PatchCollection(polys, **kw_hole))


def render_poly(poly: shapely.geometry.Polygon, ax: plt.Axes, kw=None):
    """Render an individual shapely shapely.geometry.Polygon.

    Args:
        poly (shapely.geometry.Polygon): Poly or multipoly to render.
        ax (plt.Axes): Matplotlib axis to render to.
        kw (dict): Dictionary of kwargs for the plotting.  Defaults to None.

    Returns:
        patch: ax.add_patch result
    """
    # TODO: maybe if done in batch we can speed this up?
    kw = kw or {}
    return ax.add_patch(PolygonPatch(poly, **{**style_config.poly, **kw}))


def render(
    components,
    ax=None,
    kw=None,
    # plot_format=True,
    labels=None,
    __depth=-1,  # how many sublists in we are
    _iteration=0):  # how many components we have plotted
    """Main plotting function. Plots onto an axis.

    Args:
        components: A list, dict, tuple, itterable, or shapely object.
        ax (plt.Axes): Matplotlib axis to render to.  Defaults to None.
        kw (dict): Dictionary of kwargs for the plotting.  Defaults to None.
        labels (str): Labels.  Defaults to None.
        __depth (int): How many sublists in we are.  Defaults to -1.
        _interation (int): How many components we have plotted.  Defaults to 0.

    Return:
        int: Interation
    """
    # TODO: Update to handle plotting argument with *args and **kwargs
    # but this needs the .draw functions to be updated

    __depth = __depth + 1
    #print(__depth, _iteration, components, '\n')

    kw = kw or {}
    ax = ax or plt.gca()

    if labels is 'auto':
        labels = list(map(str, range(len(components))))

    if not labels is None:
        kw = {**dict(label=labels[_iteration]), **kw}

    # Handle itterables
    if isinstance(components, dict):
        for _, objs in components.items():
            _iteration = render(objs,
                                ax=ax,
                                kw=kw,
                                labels=labels,
                                __depth=__depth,
                                _iteration=_iteration)
            # plot_format=plot_format, , **kwargs)
        # if plot_format and (__depth == 0):
        #    style_axis_simple(ax, labels=labels)
        return _iteration

    if isinstance(components, list):
        for objs in components:
            _iteration = render(objs,
                                ax=ax,
                                kw=kw,
                                labels=labels,
                                __depth=__depth,
                                _iteration=_iteration)
            # plot_format=plot_format,, **kwargs)
        # if plot_format and (__depth == 0):
        #    style_axis_simple(ax, labels=labels)
        return _iteration

    if not isinstance(components, BaseGeometry):  # obj is None
        return _iteration + 1

    # We have now a single object to draw
    obj = components

    if isinstance(obj,
                  (shapely.geometry.MultiPolygon, shapely.geometry.Polygon)):
        render_poly(obj, ax=ax, kw=kw)  # , **kwargs)

    else:
        if isinstance(obj, shapely.geometry.MultiPoint):
            kw = {**dict(marker='o'), **kw}
            coords = list(map(lambda x: list(x.coords)[0], list(obj)))
        else:
            coords = obj.coords

        if isinstance(obj, shapely.geometry.Point):
            kw = {**dict(marker='o'), **kw}

        ax.plot(*zip(*list(coords)), **kw)

    # if plot_format and (__depth == 0):
    #    style_axis_simple(ax, labels=labels)

    return _iteration + 1


# '''
# def draw_all_objects(components, ax, func=lambda x: x, root_name='components'):
#     """
#     #  TODO: This is very outdated, remove

#     Args:
#         components {[type]} -- [description]
#         ax {[type]} -- [description]
#         func {[type]} -- [description] Defaults to {lambdax:x}
#         root_name {str} -- [description] Defaults to {'components'}
#     """

#     # logger.debug(components.keys())
#     for name, obj in components.items():
#         if isinstance(obj, dict):
#             if name.startswith('components'):
#                 #logger.debug(f'Drawing: {root_name}.{name}')
#                 # allow transformation of components
#                 render(func(obj), ax=ax)
#             else:
#                 draw_all_objects(obj, ax, root_name=root_name+'.'+name)

#         elif is_component(obj):
#             #logger.debug(f' Metal_Object: {obj}')
#             render(obj.components, ax=ax)
# '''

##########################################################################################
# Style subroutines


def style_axis_simple(ax, labels=None):
    """Style function for axis called by `render`.

    Args:
        ax (plt.Axes): Matplotlib axis to render to.  Defaults to None.
        labels (str): Label
    """
    if ax is None:
        ax = plt.gca()

    # If 'box', change the physical dimensions of the Axes.
    # If 'datalim', change the x or y data limits.
    ax.set_adjustable('datalim')
    ax.set_aspect(1)
    # ax.autoscale() # for gui

    # Labels
    if not labels is None:
        font_p = mpl.font_manager.FontProperties()
        font_p.set_size('small')
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), prop=font_p)


rcParams = plt.matplotlib.rcParams


def get_prop_cycle():
    """Get the prop cycle."""
    prop_cycler = rcParams['axes.prop_cycle']
    if prop_cycler is None and 'axes.color_cycle' in rcParams:
        clist = rcParams['axes.color_cycle']
        prop_cycler = cycler('color', clist)
    return prop_cycler


#################################################################################
# Top level plotting calls - PLOT FULL


def style_axis_standard(ax):
    """Style the axis standardly.

    Args:
        ax (plt.Axes): Matplotlib axis to render to.  Defaults to None.
    """
    #fig = ax.figure

    # Core lines
    kw = dict(c='k', lw=1, zorder=-1, alpha=0.5)
    ax.axhline(0, **kw)
    ax.axvline(0, **kw)

    # Grid
    kw = dict(
        color='#CCCCCC',
        #zorder = -100,
        #alpha = 0.8,
        # fillstyle='left'
        # markevery=(1,1),
        # sketch_params=1
    )
    if 0:  # fix tick spacing
        loc = mpl.ticker.MultipleLocator(base=0.1)
        ax.xaxis.set_major_locator(loc)
        ax.yaxis.set_major_locator(loc)

    ax.grid(which='major', linestyle='--', **kw)
    ax.grid(which='minor', linestyle=':', **kw)
    ax.set_axisbelow(True)

    # Reset color cycle
    # ax.set_prop_cycle(None)
    ax.set_prop_cycle(color=['#91aecf', '#a3bbd6', '#6e94be', '#a3bbd6'])

    # fig.canvas.draw()
    # fig.canvas.flush_events()


def figure_spawn(fig_kw=None):
    """Spawn a simple, interactive matplotlib figure and gui with styled axis.
    Use mouse wheel and click and drag to navigate.

    Results in a plot that spawned up as a window! Make sure
    to enable non-inline matplotlib.

    ..code-block python

        %matplotlib qt
        fig_draw, ax_draw = figure_spawn()

    Args:
        fig_kw (fig_kw): fig_kw.  Defaults to None.

    Returns:
        (fig_draw, ax_draw)
    """
    if not fig_kw:
        fig_kw = {}
    fig_draw = figure_pz(**{**dict(num=1), **fig_kw})
    fig_draw.clf()

    ax_draw = fig_draw.add_subplot(1, 1, 1)
    # ax_draw.set_title('Layout')
    # If 'box', change the physical dimensions of the Axes. If 'datalim',
    # change the x or y data limits.
    ax_draw.set_adjustable('datalim')
    ax_draw.set_aspect(1)

    ax_draw.set_xlabel('X position (mm)')
    ax_draw.set_ylabel('Y position (mm)')

    # pylint: disable=import-outside-toplevel
    def clear_me():
        plt.sca(ax_draw)
        #from pyEPR.toolbox_plotting import plt_cla
        # plt_cla(ax_draw)

    ax_draw.clear_me = clear_me

    def restyle(display=True):
        style_axis_standard(ax_draw)
        from IPython.display import display
        display(fig_draw)

    ax_draw.restyle_me = restyle

    style_axis_standard(ax_draw)

    fig_draw.tight_layout()

    return fig_draw, ax_draw


#######################################################################


def _axis_set_watermark_img(ax: plt.Axes, file: str, size: float = 0.25):
    """Burn the axis watermark into the image.

    Args:
        ax (plt.Axes): Matplotlib axis to render to.  Defaults to None.
        file (str): The file.
        size (float): The size.  Defaults to None.
    """
    # Load image
    datafile = cbook.get_sample_data(str(file), asfileobj=False)
    img = image.imread(datafile)
    # im[:, :, -1] = 0.5  # set the alpha channel

    # window aspect
    # pylint: disable=invalid-name
    b = ax.get_window_extent()
    b = abs(b.height / b.width)
    # b = ax.get_tightbbox(ax.get_figure().canvas.get_renderer())
    # b = abs(b.height/b.width)

    # ALternative: fig.figimage
    # scale image: yscale = float(img.shape[1])/img.shape[0]
    # extent: (left, right, bottom, top)
    kw = dict(interpolation='gaussian', alpha=0.05, resample=True, zorder=-200)
    ax.imshow(img,
              extent=(1 - size * b, 1, 1 - size, 1),
              transform=ax.transAxes,
              **kw)


#######################################################################


def clear_axis(ax: plt.Axes):
    """Clear all plotted objects on an axis including lines, patches, tests, tables,
    artists, images, mouseovers, child axes, legends, collections, and containers.
    See: https://matplotlib.org/devdocs/api/_as_gen/matplotlib.axes.Axes.clear.html

    Args:
        ax (plt.Axes): MPL axis to clear
    """
    ax.clear()
    # pylint: disable=protected-access



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Interdigitated Transmon.
"""

from math import sin, cos
import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core.base import QComponent

#from ... import config
#if not config.is_building_docs():
#    from qiskit_metal import is_true


class TransmonInterdigitated(QComponent):
    """
    The base "TransmonInterdigitated" inherits the "QComponent" class.

    This creates a transmon pocket with two large pads connected by a Josephson
    junction. Both pads have four interdigitated "fingers" which increase the
    capacitance of the structure. There are three coupling capacitor pads with qpins
    defined; these can be connected to other structures in a design using CPWs.

    .. image::
        TransmonInterdigitated.png

    .. meta::
        :description: Transmon Interdigitated

    Default Options:
        * pad_width: '1000um' -- width of the large rectangular pads on either side
          of the junction
        * pad_height: '300um' -- height of the large rectangular pads on either side
          of the junction
        * finger_width: '50um' -- width of the "finger" on either side of the junction
        * finger_height: '100um' -- height of the "finger" on the side of the junction
        * finger_space: '50um' -- height of the Josephson Junction (equivalently; space
          between two fingers)
        * pad_pos_x: '0um' -- the internal coordinate defining the center of the bottom
          rectangular pad
        * pad_pos_y: '0um' -- the internal coordinate defining the center of the bottom
          rectangular pad
        * comb_width: '50um' -- the width of the four interdigitated combs connected to
          either pad
        * comb_space_vert: '50um' -- the space between the edge of a comb and the edge of
          the opposite rectangular pad
        * comb_space_hor: '50um' -- the space between adjacent interdigitated comb structures
        * jj_width: '20um' -- the width of the Josephson Junction located between the two
          fingers of the device
        * cc_space: '50um' -- the space between the lower rectangular pad and the coupling
          capacitor below it
        * cc_width: '100um' -- the width of the coupling capacitor located below the bottom
          rectangular pad
        * cc_height: '100um' -- the height of the coupling capacitor located below the bottom
          rectangular pad
        * cc_topleft_space: '50um' -- the space between the upper rectangular pad and the top
          left coupling capacitor
        * cc_topleft_width: '100um' -- the width of the top left coupling capacitor pad
        * cc_topleft_height: '100um' -- the height of the top left coupling capacitor pad
        * cc_topright_space: '50um' -- the space between the upper rectangular pad and the
          top right coupling capacitor
        * cc_topright_width: '100um' -- the width of the top right coupling capacitor pad
        * cc_topright_height: '100um' -- the height of the top right coupling capacitor pad
        * rotation_top_pad: '180' -- internal coordinate defining the angle of rotation
          between top and bottom pads
        * inductor_width: '20.0um' -- the width of the Josephson Junction
    """

    # Default drawing options
    default_options = Dict(pad_width='1000um',
                           pad_height='300um',
                           finger_width='50um',
                           finger_height='100um',
                           finger_space='50um',
                           pad_pos_x='0um',
                           pad_pos_y='0um',
                           comb_width='50um',
                           comb_space_vert='50um',
                           comb_space_hor='50um',
                           jj_width='20um',
                           cc_space='50um',
                           cc_width='100um',
                           cc_height='100um',
                           cc_topleft_space='50um',
                           cc_topleft_width='100um',
                           cc_topleft_height='100um',
                           cc_topright_space='50um',
                           cc_topright_width='100um',
                           cc_topright_height='100um',
                           rotation_top_pad='180',
                           inductor_width='20um')
    """Default drawing options"""

    # Name prefix of component, if user doesn't provide name
    component_metadata = Dict(short_name='component')
    """Component metadata"""

    def make(self):
        """Convert self.options into QGeometry."""

        p = self.parse_options()  # Parse the string options into numbers

        # draw the lower pad as a rectangle
        pad_lower = draw.rectangle(p.pad_width, p.pad_height, p.pad_pos_x,
                                   p.pad_pos_y)

        # draw the lower finger as a rectangle
        finger_lower = draw.rectangle(
            p.finger_width, p.finger_height, p.pad_pos_x, p.pad_pos_y +
            0.49999 * (p.pad_height) + 0.49999 * (p.finger_height))

        # draw the Josephson Junction as a LineString
        rect_jj = draw.LineString([
            (0, 0.5 * p.pad_height + p.finger_height),
            (0, 0.5 * p.pad_height + p.finger_height + p.finger_space)
        ])

        # draw the first comb to the right of the lower finger as a rectangle
        comb1_lower = draw.rectangle(
            p.comb_width, (2 * p.finger_height),
            (0.5 * p.finger_width + p.comb_space_hor + 0.5 * p.comb_width),
            (p.pad_pos_y + 0.5 * p.pad_height + 1.0 * p.finger_height))

        # draw the second comb to the right of the lower finger by translating the first comb
        comb2_lower = draw.translate(comb1_lower,
                                     2.0 * (p.comb_space_hor + p.comb_width),
                                     0.0)

        # draw the first comb to the left of the lower finger
        comb3_lower = draw.rectangle(
            p.comb_width, (2 * p.finger_height),
            (-0.5 * p.finger_width - 2.0 * p.comb_space_hor -
             1.5 * p.comb_width),
            (p.pad_pos_y + 0.5 * p.pad_height + 1.0 * p.finger_height))

        # draw the second comb to the left of the lower finger
        comb4_lower = draw.translate(comb3_lower,
                                     -2.0 * (p.comb_space_hor + p.comb_width),
                                     0.0)

        coupling_capacitor = draw.rectangle(
            p.cc_width, p.cc_height, p.pad_pos_x,
            p.pad_pos_y - 0.5 * (p.pad_height) - p.cc_space - 0.5 * p.cc_height)

        cc_topleft = draw.rectangle(
            p.cc_topleft_width, p.cc_topleft_height,
            p.pad_pos_x - 0.5 * p.pad_width + 0.5 * p.cc_topleft_width,
            p.pad_pos_y + 1.5 * p.pad_height + 2.0 * p.finger_height +
            p.finger_space + p.cc_topleft_space + 0.5 * p.cc_topleft_height)

        cc_topright = draw.translate(
            cc_topleft,
            p.pad_width - 0.5 * p.cc_topleft_width - 0.5 * p.cc_topright_width,
            0.0)

        # merge the bottom elements
        bottom = draw.union(pad_lower, finger_lower, comb1_lower, comb2_lower,
                            comb3_lower, comb4_lower)

        # create the top portion of the comb by translating and rotating
        # the bottom portion of the comb
        top = draw.translate(bottom, 0.0, p.pad_height + p.finger_space)
        top = draw.rotate(top, p.rotation_top_pad)

        # draw the transmon pocket bounding box
        pocket = draw.rectangle(1.5 * p.pad_width, 5.0 * p.pad_height)

        # the origin is originally set to the middle of the lower pad.
        # Let's move it to the center of the JJ.
        bottom = draw.translate(
            bottom, 0.0,
            -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)
        top = draw.translate(
            top, 0.0,
            -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)
        coupling_capacitor = draw.translate(
            coupling_capacitor, 0.0,
            -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)
        cc_topleft = draw.translate(
            cc_topleft, 0.0,
            -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)
        cc_topright = draw.translate(
            cc_topright, 0.0,
            -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)
        rect_jj = draw.translate(
            rect_jj, 0.0,
            -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)

        # now translate the final structure according to the user input
        bottom = draw.rotate(bottom, p.orientation, origin=(0, 0))
        bottom = draw.translate(bottom, p.pos_x, p.pos_y)
        top = draw.rotate(top, p.orientation, origin=(0, 0))
        top = draw.translate(top, p.pos_x, p.pos_y)
        coupling_capacitor = draw.rotate(coupling_capacitor,
                                         p.orientation,
                                         origin=(0, 0))
        coupling_capacitor = draw.translate(coupling_capacitor, p.pos_x,
                                            p.pos_y)
        cc_topleft = draw.rotate(cc_topleft, p.orientation, origin=(0, 0))
        cc_topleft = draw.translate(cc_topleft, p.pos_x, p.pos_y)
        cc_topright = draw.rotate(cc_topright, p.orientation, origin=(0, 0))
        cc_topright = draw.translate(cc_topright, p.pos_x, p.pos_y)
        rect_jj = draw.rotate(rect_jj, p.orientation, origin=(0, 0))
        rect_jj = draw.translate(rect_jj, p.pos_x, p.pos_y)
        pocket = draw.rotate(pocket, p.orientation, origin=(0, 0))
        pocket = draw.translate(pocket, p.pos_x, p.pos_y)

        # add each shape separately
        geom1 = {'pad_bot': bottom}
        geom2 = {'pad_top': top}
        geom3 = {'readout': coupling_capacitor}
        geom4 = {'bus1': cc_topleft}
        geom5 = {'bus2': cc_topright}
        geom_pocket = {'pocket': pocket}
        geom_jj = {'design': rect_jj}

        # add to qgeometry
        self.add_qgeometry('poly', geom1, layer=p.layer, subtract=False)
        self.add_qgeometry('poly', geom2, layer=p.layer, subtract=False)
        self.add_qgeometry('poly', geom3, layer=p.layer, subtract=False)
        self.add_qgeometry('poly', geom4, layer=p.layer, subtract=False)
        self.add_qgeometry('poly', geom5, layer=p.layer, subtract=False)

        self.add_qgeometry('poly', geom_pocket, layer=p.layer, subtract=True)
        self.add_qgeometry('junction',
                           geom_jj,
                           layer=p.layer,
                           subtract=False,
                           width=p.inductor_width)

        ###################################################################

        # Add Qpin connections for coupling capacitors

        # define a function that both rotates and translates the
        # qpin coordinates
        def qpin_rotate_translate(x):
            """ This function rotates the coordinates of the three qpins
            according to the user inputs for "pos_x", "pos_y"
            and "orientation".
            """
            y = list(x)
            z = [0.0, 0.0]
            z[0] = y[0] * cos(p.orientation * 3.14159 / 180) - y[1] * sin(
                p.orientation * 3.14159 / 180)
            z[1] = y[0] * sin(p.orientation * 3.14159 / 180) + y[1] * cos(
                p.orientation * 3.14159 / 180)
            z[0] = z[0] + p.pos_x
            z[1] = z[1] + p.pos_y
            x = (z[0], z[1])
            return x

        # Add Qpin connections for the bottom coupling capacitor

        qp1a = (0.0,
                -0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)
        qp1b = (0.0, -0.5 * p.pad_height - p.cc_space - p.cc_height -
                0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)

        # rotate and translate the qpin coordinates
        qp1a = qpin_rotate_translate(qp1a)
        qp1b = qpin_rotate_translate(qp1b)

        self.add_pin('readout',
                     points=np.array([qp1a, qp1b]),
                     width=0.01,
                     input_as_norm=True)

        # Add Qpin connections for top left coupling capacitor

        qp2a = (p.pad_pos_x - 0.5 * p.pad_width + 0.5 * p.cc_topleft_width,
                p.pad_pos_y + 1.5 * p.pad_height + 2.0 * p.finger_height +
                p.finger_space + p.cc_topleft_space +
                0.5 * p.cc_topleft_height - 0.5 * p.pad_height -
                p.finger_height - 0.5 * p.finger_space)
        qp2b = (p.pad_pos_x - 0.5 * p.pad_width, p.pad_pos_y +
                1.5 * p.pad_height + 2.0 * p.finger_height + p.finger_space +
                p.cc_topleft_space + 0.5 * p.cc_topleft_height -
                0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)

        qp2a = qpin_rotate_translate(qp2a)
        qp2b = qpin_rotate_translate(qp2b)

        self.add_pin('bus1',
                     points=np.array([qp2a, qp2b]),
                     width=0.01,
                     input_as_norm=True)

        # Add Qpin connections for top right coupling capacitor

        qp3a = (p.pad_pos_x + 0.5 * p.pad_width - 0.5 * p.cc_topleft_width,
                p.pad_pos_y + 1.5 * p.pad_height + 2.0 * p.finger_height +
                p.finger_space + p.cc_topleft_space +
                0.5 * p.cc_topleft_height - 0.5 * p.pad_height -
                p.finger_height - 0.5 * p.finger_space)
        qp3b = (p.pad_pos_x + 0.5 * p.pad_width, p.pad_pos_y +
                1.5 * p.pad_height + 2.0 * p.finger_height + p.finger_space +
                p.cc_topleft_space + 0.5 * p.cc_topleft_height -
                0.5 * p.pad_height - p.finger_height - 0.5 * p.finger_space)

        qp3a = qpin_rotate_translate(qp3a)
        qp3b = qpin_rotate_translate(qp3b)

        self.add_pin('bus2',
                     points=np.array([qp3a, qp3b]),
                     width=0.01,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=broad-except
# pylint: disable-msg=too-many-public-methods
# pylint: disable-msg=import-error
"""Qiskit Metal unit tests for speed."""

import unittest
import time
from qiskit_metal.tests.custom_decorators import timeout


class TestSpeed(unittest.TestCase):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    @timeout(5)
    def test_example_test(self):
        """
        Example test - Play with me to get comfortable with @timeout.
        """
        time.sleep(4)
        self.assertEqual(4, 2 + 2)


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""File contains dictionary for NSquareSpiral and the make()."""

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent
import numpy as np


class NSquareSpiral(QComponent):
    """A n count square spiral.

    Inherits `QComponent` class.

    Description:
        A n count square spiral.
        ::

            ____________
            |   _____   |
            |  |     |  |
            |  |     |  |
            |  |________|
            |

    .. image::
        NSquareSpiral.png

    .. meta::
        :description: N Square Spiral

    Default Options:
        Convention: Values (unless noted) are strings with units included,
        (e.g., '30um')

        * n: '3' -- Number of turns of the spiral
        * width: '1um' -- the width of the line of the spiral
        * radius: '40um' -- The 'radius' of the inner portion of the spiral
        * gap: '4um' -- The distance between each layer of the spiral
        * subtract: 'False'
        * helper: 'False'
    """

    default_options = Dict(n='3',
                           width='1um',
                           radius='40um',
                           gap='4um',
                           subtract='False',
                           helper='False')
    """Default drawing options"""

    TOOLTIP = """An n count square spiral"""

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc."""
        p = self.p  # p for parsed parameters. Access to the parsed options.
        n = int(p.n)
        # Create the geometry

        spiral_list = []

        for step in range(n):
            point_value = p.radius / 2 + step * (p.width + p.gap)
            spiral_list.append((-point_value, -point_value))
            spiral_list.append((point_value, -point_value))
            spiral_list.append((point_value, point_value))
            spiral_list.append((-point_value - (p.width + p.gap), point_value))

        point_value = p.radius / 2 + (step + 1) * (p.width + p.gap)
        spiral_list.append((-point_value, -point_value))
        spiral_list = draw.LineString(spiral_list)

        spiral_list = draw.rotate(spiral_list, p.orientation, origin=(0, 0))
        spiral_list = draw.translate(spiral_list, p.pos_x, p.pos_y)

        ##############################################
        # add qgeometry
        self.add_qgeometry('path', {'n_spiral': spiral_list},
                           width=p.width,
                           subtract=p.subtract,
                           helper=p.helper,
                           layer=p.layer,
                           chip=p.chip)

        points = np.array(spiral_list.coords)
        # FIX POINTS,
        self.add_pin('spiralPin',
                     points=points[-2:],
                     width=p.width,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# WILL NEED UPDATING TO NEW ELEMENT SCHEME
"""Main module for basic geometric manipulation of non-shapely objects, but
objects such as points and arrays used in drawing."""

import math
from collections.abc import Iterable, Mapping
from typing import List, Tuple, Union

import numpy as np
import shapely
import shapely.wkt
from numpy import array
from numpy.linalg import norm
from shapely.geometry import MultiPolygon, Polygon

from .. import logger
from .. import is_component
from . import BaseGeometry

__all__ = [
    'get_poly_pts', 'get_all_component_bounds', 'get_all_geoms',
    'flatten_all_filter', 'remove_colinear_pts', 'array_chop',
    'vec_unit_planar', 'Vector'
]

# Used for numpy.round()
PRECISION = 10

#########################################################################
# Shapely Geometry Basic Coordinates


def get_poly_pts(poly: Polygon):
    """Return the coordinates of a Shapely polygon with the last repeating
    point removed.

    Args:
        poly (shapely.Polygon): Shapely polygin

    Returns:
        np.array: Sequence of coordinates.
    """
    return np.array(poly.exterior.coords)[:-1]


def get_all_geoms(obj, func=lambda x: x, root_name='components'):
    """Get a dict of dict of all shapely objects, from components, dict, etc.

    Used to compute the bounding box.

    Args:
        obj (dict, list, element, component): Object to get from.
        func (function): Function to use if mapping. Defaults to (lambda x: x).
        root_name (str): Name to prepend in the flattening. Defaults to 'components'.

    Returns:
        dict: Dictionary of geometries
    """

    # Prelim
    # Calculate the new name
    def new_name(name):        return root_name + '.' + \
name if not (root_name == '') else name

    # Check what we have

    if is_component(obj):
        # Is it a metal component? Then traverse its components
        return obj.get_all_geom()  # dict of shapely geoms

    # elif is_element(obj):
    #    # is it a metal element?
    #    return {obj.name: obj.geom}  # shapely geom

    elif isinstance(obj, BaseGeometry):
        # Is it the shapely object? This is the bottom of the search tree, then return
        return obj

    elif isinstance(obj, Mapping):
        return {
            get_all_geoms(sub_obj, root_name=new_name(name))
            for name, sub_obj in obj.items()
        }
        '''
        RES = {}
        for name, sub_obj in obj.items():
            if is_component(obj):
                RES[name] = get_all_geoms(
                    sub_obj.components, root_name=new_name(name))
            elif isinstance(sub_obj, dict):
                # if name.startswith('components'): # old school to remove eventually TODO
                #    RES[name] = func(obj) # allow transformaiton of components
                # else:
                RES[name] = get_all_geoms(sub_obj, root_name=new_name(name))
            elif isinstance(sub_obj, BaseGeometry):
                RES[name] = func(obj)
        return RES
        '''

    else:
        logger.debug(
            f'warning: {root_name} was not an object or dict or the right handle'
        )
        return None


def flatten_all_filter(components: dict, filter_obj=None):
    """Internal function to flatten a dict of shapely objects.

    Args:
        components (dict): Dictionary of components
        filter_obj (class): Filter based on this class.  Defaults to None.

    Returns:
        array: Flattened dictionary
    """
    assert isinstance(components, dict)

    RES = []
    for name, obj in components.items():
        if isinstance(obj, dict):
            RES += flatten_all_filter(obj, filter_obj)
        else:
            if filter_obj is None:
                RES += [obj]  # add whatever we have in here
            else:
                if isinstance(obj, filter_obj):
                    RES += [obj]
                else:
                    print('flatten_all_filter: ', name)

    return RES


def get_all_component_bounds(components: dict, filter_obj=Polygon):
    """Pass in a dict of components to calculate the total bounding box.

    Args:
        components (dict): Dictionary of components
        filter_obj (Polygon): Only use instances of this object to
                              calculate the bounds

    Returns:
        tuple: (x_min, y_min, x_max, y_max)
    """
    assert isinstance(components, dict)

    components = get_all_geoms(components)
    components = flatten_all_filter(components, filter_obj=filter_obj)

    (x_min, y_min, x_max, y_max) = MultiPolygon(components).bounds

    return (x_min, y_min, x_max, y_max)


def round_coordinate_sequence(geom_ref, precision):
    """Rounds the vertices of a coordinate sequence (both interior and
    exterior).

    Args:
        geometry (shapely.geometry) : A shapely geometry, should not be a MultiPoly
        precision (int) : The decimal precision to round to (eg. 3 -> 0.001)
    Returns:
        shapely.geometry : A shapely geometry with rounded coordinates
    """

    new_geom_ref = shapely.wkt.loads(
        shapely.wkt.dumps(geom_ref, rounding_precision=precision))

    return new_geom_ref


#########################################################################
# POINT LIST FUNCTIONS


def check_duplicate_list(your_list):
    """Check if the list contains duplicates.

    Args:
        your_list (list): List to check

    Returns:
        bool: True if there are duplicates, False otherwise
    """
    return len(your_list) != len(set(your_list))


def array_chop(vec, zero=0, rtol=0, machine_tol=100):
    """Chop array entries close to zero.

    Args:
        vec (array): Array to chop
        zero (double): Value to check against.  Defaults to 0.
        rtol (double): Relative tolerance.  Defaults to 0.
        machine_tol (double): Machine tolerance.  Defaults to 100.

    Returns:
        array: Chopped array
    """
    vec = np.array(vec)
    mask = np.isclose(vec,
                      zero,
                      rtol=rtol,
                      atol=machine_tol * np.finfo(float).eps)
    vec[mask] = 0
    return vec


def remove_colinear_pts(points):
    """Remove colinear points and identical consecqutive points.

    Args:
        points (array): Array of points

    Returns:
        ndarray: A copy of the input array without colinear points
    """
    remove_idx = []
    for i in range(2, len(points)):
        v1 = array(points[i - 2]) - array(points[i - 1])
        v2 = array(points[i - 1]) - array(points[i - 0])
        if Vector.are_same(v1, v2):
            remove_idx += [i - 1]
        elif Vector.angle_between(v1, v2) == 0:
            remove_idx += [i - 1]
    points = np.delete(points, remove_idx, axis=0)

    # remove  consecutive duplicates
    remove_idx = []
    for i in range(1, len(points)):
        if norm(points[i] - points[i - 1]) == 0:
            remove_idx += [i]
    points = np.delete(points, remove_idx, axis=0)

    return points


#########################################################################
# Points, Lines and Areas functions
def intersect(p1x, p1y, p2x, p2y, x0, y0):
    """Intersect segment defined by p1 and p2 with ray coming out of x0,y0 ray
    can be horizontal y=y0  x=x0+dx , want dx>0.

    Args:
        p1x (float): x coordinate of point 1 of segment
        p1y (float): y coordinate of point 1 of segment
        p2x (float): x coordinate of point 2 of segment
        p2y (float): y coordinate of point 2 of segment
        x0 (float): x coordinate anchoring the intersection ray
        y0 (float): y coordinate anchoring the intersection ray

    Returns:
        boolean int: (1) if intersecting, (0) if not intersecting
    """
    if p1x != p2x and p1y != p2y:
        m = (p2y - p1y) / (p2x - p1x)
        x_inter = (y0 - p1y) / m + p1x
        if x_inter >= x0 and np.min([p1y, p2y]) <= y0 <= np.max([p1y, p2y]):
            ans = 1
        else:
            ans = 0
    else:
        if p1x == p2x:  # vertical segment
            if x0 <= p1x and np.min([p1y, p2y]) <= y0 <= np.max([p1y, p2y]):
                ans = 1
            else:
                ans = 0
        if p1y == p2y:  # horizontal segment
            if y0 == p1y:
                ans = 1
            else:
                ans = 0
    return ans


def in_or_out(xs, ys, x0, y0):
    """Count up how many times a ray intersects the polygon, even or odd tells
    you whether inside (odd) or outside (even)"""
    crossings = 0
    for i in range(len(xs) - 1):
        p1x = xs[i]
        p2x = xs[i + 1]
        p1y = ys[i]
        p2y = ys[i + 1]
        cross = intersect(p1x, p1y, p2x, p2y, x0, y0)
        # print('i = ', i, 'cross = ', cross)
        crossings += cross
    return crossings


#########################################################################
# Vector functions


def vec_unit_planar(vector: np.array):
    """Make the planar 2D (x,y) part of a vector to be unit mag. Return a
    vector where is XY components now a unit vector. I.e., Normalizes only in
    the XY plane, leaves the Z plane alone.

    Args:
        vector (np.array): Input 2D or 3D

    Returns:
        np.array: Same dimension 2D or 3D

    Raises:
        Exception: The input was not a 2 or 3 vector
    """
    vector = array_chop(vector)  # get rid of near zero crap

    if len(vector) == 2:
        _norm = norm(vector)

        if not bool(_norm):  # zero length vector
            logger.debug(f'Warning: zero vector length')
            return vector

        return vector / _norm

    elif len(vector) == 3:
        v2 = vec_unit_planar(vector[:2])
        return np.append(v2, vector[2])

    else:
        raise Exception('You did not give a 2 or 3 vec')


# pylint: disable=invalid-name
def to_vec3D(list_of_2d_pts: List[Tuple], z=0) -> np.ndarray:
    """Adds 3rd point to list of 2D points. For the given design, get the z
    values in HFSS UNITS! Manually specify z dimension.

    Args:
        list_of_2d_pts (List[Tuple]): List of 2D points
        z (int, optional): z-value in hfss. Defaults to 0.

    Returns:
        np.ndarray: vec3d of points
    """
    add_me = [z]

    vec3d = np.array([list(a_2d_pt) + add_me for a_2d_pt in list_of_2d_pts],
                     dtype="object")
    return vec3d


def to_vec3D_list(list_of_2d_pts: List[Tuple], z=0) -> list:
    """Adds 3rd point to list of 2D points. For the given design, get the z
    values in HFSS UNITS! Manually specify z dimension.

    Args:
        list_of_2d_pts (List[Tuple]): List of 2D points
        z (int, optional): z-value in hfss. Defaults to 0.

    Returns:
        list: vec3d of points
    """
    add_me = [z]

    vec3d_list = [list(a_2d_pt) + add_me for a_2d_pt in list_of_2d_pts]
    return vec3d_list


Vec2D = Union[list, np.ndarray]


class Vector:
    """Utility functions to call on 2D vectors, which can be np.ndarrays or
    lists."""

    normal_z = np.array([0, 0, 1])
    """ Noraml Z array """

    @staticmethod
    def rotate_around_point(xy: Vec2D, radians: float,
                            origin=(0, 0)) -> np.ndarray:
        r"""Rotate a point around a given point.
        Positive angles are counter-clockwise and negative are clockwise rotations.

        .. math::

            \begin{split}
            x_\mathrm{off} &= x_0 - x_0 \cos{\theta} + y_0 \sin{\theta} \\
            y_\mathrm{off} &= y_0 - x_0 \sin{\theta} - y_0 \cos{\theta}
            \end{split}

        Args:
            xy (Vec2D): A 2D vector.
            radians (float): Counter clockwise angle.
            origin (tuple): point to rotate about.  Defaults to (0, 0).

        Returns:
            np.ndarray: rotated point
        """
        # see: https://gist.github.com/LyleScott/d17e9d314fbe6fc29767d8c5c029c362
        x, y = xy
        offset_x, offset_y = origin
        adjusted_x = (x - offset_x)
        adjusted_y = (y - offset_y)
        cos_rad = math.cos(radians)
        sin_rad = math.sin(radians)
        qx = offset_x + cos_rad * adjusted_x - sin_rad * adjusted_y
        qy = offset_y + sin_rad * adjusted_x + cos_rad * adjusted_y
        return qx, qy

    @staticmethod
    def rotate(xy: Vec2D, radians: float) -> np.ndarray:
        """Counter-clockwise rotation of the vector in radians. Positive angles
        are counter-clockwise and negative are clockwise rotations.

        Args:
            xy (Vec2D): A 2D vector.
            radians (float): Counter clockwise angle

        Returns:
            np.ndarray: Rotated point
        """
        x, y = xy
        cos_rad = math.cos(radians)
        sin_rad = math.sin(radians)
        qx = cos_rad * x - sin_rad * y
        qy = sin_rad * x + cos_rad * y
        return np.array([qx, qy])  #ADD ARRAY CHOP - draw_utility

    @staticmethod
    def angle(vector: Vec2D) -> float:
        """Return the angle in radians of a vector.

        Args:
            vector (Union[list, np.ndarray): A 2D vector

        Returns:
            float: Angle in radians

        Caution:
            The angle is defined from the Y axis!

            See https://docs.scipy.org/doc/numpy/reference/generated/numpy.arctan2.html
            ::

                    |-> Positive direction, defined from Y axis
                    |
                --------
                    |
                    |

                    x1       x2       arctan2(x1,x2)
                    +/- 0    +0       +/- 0
                    +/- 0    -0       +/- pi
                    > 0      +/-inf   +0 / +pi
                    < 0      +/-inf   -0 / -pi
                    +/-inf   +inf     +/- (pi/4)
                    +/-inf   -inf     +/- (3*pi/4)

            Note that +0 and -0 are distinct floating point numbers, as are +inf and -inf.
        """
        return np.arctan2(vector)

    @staticmethod
    def angle_between(v1: Vec2D, v2: Vec2D) -> float:
        """Returns the angle in radians between vectors 'v1' and 'v2'.
        Disclaimer: The same method is also used Vec3D as it is flexible enough.

        Args:
            v1 (Vec2D): First vector
            v2 (Vec2D): Second vector

        Returns:
            float: Angle in radians. The angle of the ray intersecting the unit
            circle at the given x-coordinate in radians [0, pi]. This is a scalar.
        """
        v1_u = vec_unit_planar(v1)
        v2_u = vec_unit_planar(v2)
        return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))

    @staticmethod
    def add_z(vec2D: np.array, z: float = 0.):
        """Turn a 2D vector into a 3D vector by adding the z coorindate.

        Args:
            vec2D (np.array): Input 2D vector.
            z (float): Add this value to the 3rd dimension.  Defaults to {0}.

        Returns:
            np.array: 3D vector.
        """
        if isinstance(vec2D[0], Iterable):
            return array([Vector.add_z(vec, z=z) for vec in vec2D])
        else:
            return array(list(vec2D) + [z])

    @staticmethod
    def normed(vec: Vec2D) -> Vec2D:
        """Return normed vector.

        Args:
            vec (Vec2D): Vector

        Returns:
            vector: Vec2D -- Unit normed version of vector
        """
        return vec / norm(vec)

    @staticmethod
    def norm(vec: Vec2D) -> float:
        """Return the norm of a 2D vector.

        Args:
            vec (Vec2D): 2D vector

        Returns:
            float: Length of vector
        """
        return norm(vec)

    @staticmethod
    def are_same(v1: Vec2D, v2: Vec2D, tol: int = 100) -> bool:
        """Check if two vectors are within an infinitesimal distance set by
        `tol` and machine epsilon.
        Disclaimer: The same method is also used Vec3D as it is flexible enough.

        Args:
            v1 (Vec2D): First vector to check
            v2 (Vec2D): Second vector to check
            tol (int): How much to multiply the machine precision, np.finfo(float).eps,
                       by as the tolerance.  Defaults to 100.

        Returns:
            bool: Same or not
        """
        v1, v2 = np.array(v1), np.array(v2)
        return Vector.is_zero(v1 - v2, tol=tol)

    @staticmethod
    def is_zero(vec: Vec2D, tol: int = 100) -> bool:
        """Check if a vector is essentially zero within machine precision, set
        by `tol` and machine epsilon.

        Args:
            vec (Vec2D): Vector to check
            tol (int): How much to multiply the machine precision, np.finfo(float).eps,
                       by as the tolerance.  Defaults to 100.

        Returns:
            bool: Close to zero or not
        """
        return float(norm(vec)) < tol * np.finfo(float).eps

    @staticmethod
    def get_distance(u: Union[tuple, list, np.ndarray],
                     v: Union[tuple, list, np.ndarray],
                     precision: int = 9) -> float:
        """Get the Euclidean distance between points u and v to the specified
        precision.

        Args:
            u (Union[tuple, list, np.ndarray]): Coordinates of a point.
            v (Union[tuple, list, np.ndarray]): Coordinates of a second point.
            precision (int, optional): Precision of the result. Defaults to 9.

        Returns:
            float: Distance between u anbd v.
        """
        u, v = np.array(u), np.array(v)
        return round(abs(norm(u - v)), precision)

    @staticmethod
    def two_points_described(points2D: List[Vec2D]) -> Tuple[np.ndarray]:
        """Get the distance, units and tangents.

        Args:
            points (np.array or list): 2D list of points

        Returns:
            tuple: (distance_vec, dist_unit_vec, tangent_vec) -- Each is a vector np.array

        For a list of exactly two given 2D points, get:
            * d: the distance vector between them
            * n: normal vector defined by d
            * t: the vector tangent to n

        .. code-block:: python

            vec_D, vec_d, vec_n = Vector.difference_dnt(points)
        """
        assert len(points2D) == 2
        start = np.array(points2D[0])
        end = np.array(points2D[1])

        distance_vec = end - start  # distance vector
        # unit vector along the direction of the two point
        unit_vec = distance_vec / norm(distance_vec)
        # tangent vector counter-clockwise 90 deg rotation
        tangent_vec = np.round(Vector.rotate(unit_vec, np.pi / 2),
                               decimals=PRECISION)

        if Vector.is_zero(distance_vec):
            logger.debug(
                f'Function `two_points_described` encountered a zero vector'
                ' length. The two points should not be the same.')

        return distance_vec, unit_vec, tangent_vec

    @staticmethod
    def snap_unit_vector(vec_n: Vec2D, flip: bool = False) -> Vec2D:
        """snaps to either the x or y unit vectors.

        Args:
            vec_n (Vec2D): 2D vector
            flip (bool): True to flip.  Defaults to False.

        Returns:
            Vec2D: Snapped vector
        """
        #TODO: done silly, fix up
        m = np.argmax(abs(vec_n))
        m = m if flip is False else int(not m)
        v = np.array([0, 0])
        v[m] = np.sign(vec_n[m])
        vec_n = v
        return vec_n


class Vec3D(Vector):
    """Class to define a 3D vector and perform operations on them.

    Returns:
        Vec3D: A 3-dimensional vector object
    """

    @staticmethod
    def norm(vec: np.ndarray) -> float:
        """Magnitude of the vector

        Args:
            vec (np.ndarray): input vector

        Returns:
            float: magnitude of the vector calculated by Frobenius norm
        """
        return np.round(np.linalg.norm(vec), decimals=PRECISION)

    @staticmethod
    def normed(vec: np.ndarray):
        """Normed vector

        Args:
            vec (np.ndarray): input vector

        Returns:
            np.ndarray: normed vector
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        return vec / Vec3D.norm(vec)

    @staticmethod
    def add(vec: np.ndarray, other: np.ndarray):
        """Adds two vectors

        Args:
            vec (np.ndarray): input vector
            other (np.ndarray): the other vector

        Returns:
            np.ndarray: resultant vector
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        if not isinstance(other, np.ndarray):
            other = np.array(other)
        return (vec + other)

    @staticmethod
    def dot(vec: np.ndarray, other: np.ndarray) -> float:
        """Dot product of two vectors

        Args:
            vec (np.ndarray): input vector
            other (np.ndarray): the other vector

        Returns:
            float: result from dot product
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        if not isinstance(other, np.ndarray):
            other = np.array(other)
        return np.round(vec.dot(other), decimals=PRECISION)

    @staticmethod
    def cross(vec: np.ndarray, other: np.ndarray):
        """Cross product of two vectors

        Args:
            vec (np.ndarray): input vector
            other (np.ndarray): the other vector

        Returns:
            np.ndarray: resultant vector
        """
        return np.round(np.cross(vec, other), decimals=PRECISION)

    @staticmethod
    def sub(vec: np.ndarray, other: np.ndarray):
        """Difference between two vectors

        Args:
            vec (np.ndarray): input vector
            other (np.ndarray): the other vector

        Returns:
            np.ndarray: resultant vector
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        if not isinstance(other, np.ndarray):
            other = np.array(other)
        return (vec - other)

    @staticmethod
    def scale(vec: np.ndarray, value: Union[int, float]):
        """Multiply the vector with a scalar

        Args:
            vec (np.ndarray): input vector
            value (Union[int, float]): scalar value

        Returns:
            np.ndarray: resultant vector
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        return (vec * value)

    @staticmethod
    def get_distance(vec: np.ndarray, other: np.ndarray) -> float:
        """Calculate Euler distance between two vectors

        Args:
            vec (np.ndarray): input vector
            other (np.ndarray): the other vector

        Returns:
            float: Euler distance
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        if not isinstance(other, np.ndarray):
            other = np.array(other)
        return np.round(Vec3D.norm(vec - other), decimals=PRECISION)

    @staticmethod
    def translate(vec: np.ndarray, translate_vec: np.ndarray):
        """Translate a vector

        Args:
            vec (np.ndarray): vector to transform
            translate_vec (np.ndarray): translation values as a vector

        Returns:
            np.ndarray: return a new np.ndarray object
        """
        return Vec3D.add(vec, translate_vec)

    @staticmethod
    def rotate(vec: np.ndarray,
               cvec: np.ndarray,
               ax: bool = False,
               ay: bool = False,
               az: bool = False,
               radians: float = 0.0):
        """Rotate a vector

        Args:
            vec (np.ndarray): vector to transform
            cvec (np.ndarray): center of rotation as a vector
            ax (bool, optional): Rotate around x-axis. Defaults to False.
            ay (bool, optional): Rotate around y-axis. Defaults to False.
            az (bool, optional): Rotate around z-axis. Defaults to False.
            radians (float, optional): radians of rotation. Defaults to 0.0.

        Returns:
            np.ndarray: return a new np.ndarray object
        """
        if not isinstance(vec, np.ndarray):
            vec = np.array(vec)
        if not isinstance(cvec, np.ndarray):
            cvec = np.array(cvec)

        if ax:
            rot_x = np.array([[1, 0, 0], [0,
                                          np.cos(radians), -np.sin(radians)],
                              [0, np.sin(radians),
                               np.cos(radians)]])
            translate_vec = np.array([0, cvec[1], cvec[2]])
            new_vec = rot_x.dot((vec - translate_vec)) + translate_vec

        if ay:
            rot_y = np.array([[np.cos(radians), 0,
                               np.sin(radians)], [0, 1, 0],
                              [-np.sin(radians), 0,
                               np.cos(radians)]])
            translate_vec = np.array([cvec[0], 0, cvec[2]])
            new_vec = rot_y.dot((vec - translate_vec)) + translate_vec

        if az:
            rot_z = np.array([[np.cos(radians), -np.sin(radians), 0],
                              [np.sin(radians),
                               np.cos(radians), 0], [0, 0, 1]])
            translate_vec = np.array([cvec[0], cvec[1], 0])
            new_vec = rot_z.dot((vec - translate_vec)) + translate_vec

        return new_vec

    @staticmethod
    def angle_elevation(vec: np.ndarray) -> float:
        """Returns the elevation angle of vector

        Args:
            vec (np.ndarray): input vector

        Returns:
            float: elevation angle in radians
        """
        unit_z = np.array([0, 0, 1])
        unit_self = Vec3D.normed(vec)
        return np.arccos(unit_self.dot(unit_z))

    @staticmethod
    def angle_azimuth(vec: np.ndarray) -> float:
        """Returns the azimuthal angle of vector

        Args:
            vec (np.ndarray): input vector

        Returns:
            float: azimuthal angle of vector
        """
        return np.arctan2(*vec[:2][::-1])

    @staticmethod
    def snap_unit_vector(vec: np.ndarray, snap_to: str = None) -> Vec2D:
        """Snaps to either the x, y or z unit vectors with sign.

        Args:
            vec_n (np.ndarray): 3D vector
            snap_to (str): Snap to direction with max projection length by default.
                            Snap to particular axis if specified.

        Returns:
            np.ndarray: Snapped vector
        """
        if snap_to is None:
            idx = np.argmax(np.abs(vec))
        else:
            valid_snap_to = ["x", "y", "z"]
            if snap_to not in valid_snap_to:
                raise ValueError(
                    "The argument 'snap_to' should be one of 'x', 'y', or 'z'.")
            idx = valid_snap_to.index(snap_to)

        new_vec = np.array([0, 0, 0])
        new_vec[idx] = np.sign(vec[idx])
        return new_vec

    @staticmethod
    def two_points_described(points3D: List[np.ndarray],
                             user_plane: np.ndarray) -> Tuple[np.ndarray]:
        """Get the distance, units and tangents.

        Args:
            points3D (List[np.ndarray]): 3D list of points
            user_plane (np.ndarray): 2D plane containing the input points and
                                    defined by coefficients A,B,C,D in
                                    equation (Ax + By + Cz + D = 0)

        Returns:
            tuple: (distance_vec, dist_unit_vec, tangent_vec) -- Each is a vector np.array
        """
        if len(points3D) != 2:
            raise ValueError(f"Expected only 2 points, found {len(points3D)}.")
        if len(user_plane) != 4:
            raise ValueError(
                f"""Expected all four coefficients A,B,C,D for the plane.
                Found {len(user_plane)} instead.""")

        if not isinstance(user_plane, np.ndarray):
            user_plane = np.array(user_plane)
        if not isinstance(points3D, np.ndarray):
            points3D = np.array(points3D)

        # Check if both points are on the user_plane
        pt1_on_plane = True if (np.round(user_plane.dot(
            np.append(points3D[0], [1])),
                                         decimals=PRECISION) == 0) else False
        pt2_on_plane = True if (np.round(user_plane.dot(
            np.append(points3D[1], [1])),
                                         decimals=PRECISION) == 0) else False

        if pt1_on_plane and pt2_on_plane:
            distance_vec = points3D[0] - points3D[1]
            dist_unit_vec = Vec3D.normed(distance_vec)
            plane_normal = Vec3D.normed(user_plane[:3])
            tangent_vec = Vec3D.cross(dist_unit_vec, plane_normal)
            return distance_vec, dist_unit_vec, tangent_vec
        else:
            raise ValueError("Either one or both points not on the user_plane.")

    @staticmethod
    def add_z(vec2D: np.array, z: float = 0.):
        raise NotImplementedError(
            "This method does not need to be implemented for 3D.")

    @staticmethod
    def angle(vector: Vec2D) -> float:
        raise NotImplementedError(
            """This method is implemented by 'angle_azimuth' method for 3D vector.
            Please use that instead.""")

    @staticmethod
    def rotate_around_point(xy: Vec2D, radians: float,
                            origin=(0, 0)) -> np.ndarray:
        raise NotImplementedError(
            """This method is implemented by 'rotate' method for 3D vector.
            Please use that instead.""")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This is a utility module used for qt."""
# pylint: disable=invalid-name

from types import MethodType

# from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QTimer
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QDockWidget

__all__ = ['blend_colors']


def blend_colors(color1: QColor,
                 color2: QColor,
                 r: float = 0.2,
                 alpha=255) -> QColor:
    """Blend two qt colors together.

    Args:
        color1 (QColor): first color
        color2 (QColor): second color
        r (float): ratio
        alpha (int): alpha

    Returns:
        QColor: new color
    """
    color3 = QColor(color1.red() * (1 - r) + color2.red() * r,
                    color1.green() * (1 - r) + color2.green() * r,
                    color1.blue() * (1 - r) + color2.blue() * r, alpha)
    return color3


#------------------------------------------------------------------------------------------

STYLE_HIGHLIGHT_ = r"""
QWidget {
    outline: 3px solid red;
    border: 3px solid red;
}"""


def doShowHighlighWidget(self: QDockWidget, timeout=1500, style_highlight=None):
    """Highlight temporarily, raise, show the widget.
    Force resets the style at the component to None after a period.
    """
    if style_highlight is None:
        style_highlight = STYLE_HIGHLIGHT_
    self.setStyleSheet(style_highlight)
    self.show()
    self.raise_()

    def doResetStyle(self: 'QDockWidget'):
        """Reset the style of the widget."""
        self.setStyleSheet('')

    # Bind the method dynamically to the instance using MethodType
    self.doResetStyle = MethodType(doResetStyle, self)

    # self.doResetStyle = doResetStyle.__get__(self, type(self))
    # monkey patch class instance:
    # https://stackoverflow.com/questions/28127874/monkey-patching-python-an-instance-method

    QTimer.singleShot(timeout, self.doResetStyle)


### Alternative to doShowHighlighWidget:

# from PySide6.QtWidgets import QFrame, QWidget
# from PySide6 import QtCore
# obj = gui.canvas
# frame = QFrame(obj.parent())
# frame.setGeometry(obj.frameGeometry())
# frame.setFrameShape(QFrame.Box)
# frame.setLineWidth(3)
# frame.show()

# frame.setStyleSheet(r"""
#   border-radius: 10px;
#   outline: 3px solid red;
#   border: 3px solid red;
#   background-color: transparent;
#   border-image:none;
# """)
# ## the folloing make it dissapear altoheger
# # frame.setWindowFlags(QtCore.Qt.FramelessWindowHint)
# # frame.setAttribute(QtCore.Qt.WA_TranslucentBackground)

# # Alternative see:
# https://stackoverflow.com/questions/58458323/how-to-use-qt-stylesheet-to-customize-only-partial-qwidget-border

#------------------------------------------------------------------------------------------
